# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gst, GstPbutils, GLib, Gio

from re import match
from gettext import gettext as _

from lollypop.define import App
from lollypop.logger import Logger
from lollypop.utils_file import decodeUnicode, splitUnicode
from lollypop.utils import format_artist_name, get_iso_date_from_string
from lollypop.tag_frame_text import FrameTextTag
from lollypop.tag_frame_lang import FrameLangTag


class Discoverer:
    """
        Discover tags
    """

    def __init__(self):
        """
            Init tag reader
        """

        self._discoverer = GstPbutils.Discoverer.new(10 * Gst.SECOND)

    def get_info(self, uri):
        """
            Return information for file at uri
            @param uri as str
            @Exception GLib.Error
            @return GstPbutils.DiscovererInfo
        """
        info = self._discoverer.discover_uri(uri)
        return info


class TagReader:
    """
        Scanner tag reader
    """

    __STRING = ["title", "artist", "composer", "conductor",
                "musicbrainz-albumid", "musicbrainz-trackid",
                "musicbrainz-artistid", "musicbrainz-albumartistid",
                "version", "performer", "artist-sortname",
                "album-artist-sortname", "interpreted-by", "album-artist",
                "album", "genre", "lyrics", "publisher"]
    __INT = ["album-disc-number", "track-number"]
    __DOUBLE = ["beats-per-minute"]

    def __init__(self):
        """
            Init tag reader
        """
        pass

    def get_title(self, tags, filepath):
        """
            Return title for tags
            @param tags as Gst.TagList
            @param filepath as string
            @return title as string
        """
        if tags is None:
            return GLib.path_get_basename(filepath)
        title = self.__get(tags, ["title"])
        if not title:
            title = GLib.path_get_basename(filepath)
        return title

    def get_artists(self, tags):
        """
            Return artists for tags
            @param tags as Gst.TagList
            @return string like "artist1;artist2;..."
        """
        if tags is None:
            return _("Unknown")
        return self.__get(tags, ["artist"])

    def get_composers(self, tags):
        """
            Return composers for tags
            @param tags as Gst.TagList
            @return string like "composer1;composer2;..."
        """
        if tags is None:
            return _("Unknown")
        return self.__get(tags, ["composer"])

    def get_conductors(self, tags):
        """
            Return conductors for tags
            @param tags as Gst.TagList
            @return string like "conductor1;conductor2;..."
        """
        if tags is None:
            return _("Unknown")
        return self.__get(tags, ["conductor"])

    def get_mb_id(self, tags, name):
        """
            Get MusicBrainz ID
            @param tags as Gst.TagList
            @param name as str
            @return str
        """
        if tags is None or not name:
            return ""
        return self.__get(tags, ["musicbrainz-" + name])

    def get_mb_album_id(self, tags):
        """
            Get album id (musicbrainz)
            @param tags as Gst.TagList
            @return str
        """
        return self.get_mb_id(tags, 'albumid')

    def get_mb_track_id(self, tags):
        """
            Get recording id (musicbrainz)
            @param tags as Gst.TagList
            @return str
        """
        return self.get_mb_id(tags, 'trackid')

    def get_mb_artist_id(self, tags):
        """
            Get artist id (musicbrainz)
            @param tags as Gst.TagList
            @return str
        """
        return self.get_mb_id(tags, 'artistid')

    def get_mb_album_artist_id(self, tags):
        """
            Get album artist id (musicbrainz)
            @param tags as Gst.TagList
            @return str
        """
        return self.get_mb_id(tags, 'albumartistid')

    def get_version(self, tags):
        """
            Get recording version
            @param tags as Gst.TagList
            @return str
        """
        if tags is None:
            return ""
        return self.__get(tags, ["version"])

    def get_performers(self, tags):
        """
            Return performers for tags
            @param tags as Gst.TagList
            @return string like "performer1;performer2;..."
        """
        if tags is None:
            return _("Unknown")
        return self.__get(tags, ["performer"])

    def get_artist_sortnames(self, tags):
        """
            Return artist sort names
            @param tags as Gst.TagList
            @return artist sort names as "str;str"
        """
        if tags is None:
            return ""
        return self.__get(tags, ["artist-sortname"])

    def get_album_artist_sortnames(self, tags):
        """
            Return album artist sort names
            @param tags as Gst.TagList
            @return artist sort names as "str;str"
        """
        if tags is None:
            return ""
        return self.__get(tags, ["album-artist-sortname"])

    def get_remixers(self, tags):
        """
            Get remixers tag
            @param tags as Gst.TagList
            @return artist sort names as "str,str"
        """
        if tags is None:
            return _("Unknown")
        remixers = self.__get(tags, ["interpreted-by"])
        if not remixers:
            remixers = self.__get_extended(tags, ["REMIXER"])
        return remixers

    def get_album_artists(self, tags):
        """
            Return album artists for tags
            @param tags as Gst.TagList
            @return album artist as string or None
        """
        if tags is None:
            return _("Unknown")
        return self.__get(tags, ["album-artist"])

    def get_album_name(self, tags):
        """
            Return album for tags
            @param tags as Gst.TagList
            @return album name as string
        """
        if tags is None:
            return _("Unknown")
        album = self.__get(tags, ["album"])
        if not album:
            album = _("Unknown")
        return album

    def get_genres(self, tags):
        """
            Return genres for tags
            @param tags as Gst.TagList
            @return string like "genre1;genre2;..."
        """
        if tags is None:
            return _("Unknown")
        genres = self.__get(tags, ["genre"])
        if not genres:
            genres = _("Unknown")
        return genres

    def get_discname(self, tags):
        """
            Return disc name
            @param tags as Gst.TagList
            @return disc name as str
        """
        return self.__get_extended(tags, ['PART', 'DISCSUBTITLE'])

    def get_discnumber(self, tags):
        """
            Return disc number for tags
            @param tags as Gst.TagList
            @return disc number as int
        """
        if tags is None:
            return 0
        discnumber = self.__get(tags, ["album-disc-number"])
        if not discnumber:
            discnumber = 0
        return discnumber

    def get_compilation(self, tags):
        """
            Return True if album is a compilation
            @param tags as Gst.TagList
            @return bool
        """
        if tags is None:
            return False
        try:
            compilation = self.__get_private_string(tags, "TCMP", False)
            if not compilation:
                compilation = self.__get_extended(tags, ["COMPILATION"])
            if compilation:
                return bool(compilation)
        except Exception as e:
            Logger.error("TagReader::get_compilation(): %s" % e)
        return False

    def get_tracknumber(self, tags, filename):
        """
            Return track number for tags
            @param tags as Gst.TagList
            @param filename as str
            @return track number as int
        """
        if tags is not None:
            tracknumber = self.__get(tags, ["track-number"])
        else:
            tracknumber = None
        if not tracknumber:
            # Guess from filename
            m = match("^([0-9]*)[ ]*-", filename)
            if m:
                try:
                    tracknumber = int(m.group(1))
                except:
                    tracknumber = 0
            else:
                tracknumber = 0
        return min(abs(tracknumber), GLib.MAXINT32)

    def get_year(self, tags):
        """
            Return track year for tags
            @param tags as Gst.TagList
            @return year and timestamp (int, int)
        """
        try:
            (exists_date, date) = tags.get_date_index("date", 0)
            (exists_datetime, datetime) = tags.get_date_time_index("datetime",
                                                                   0)
            year = timestamp = None
            if exists_datetime:
                if datetime.has_year():
                    year = datetime.get_year()
                if datetime.has_month():
                    month = datetime.get_month()
                else:
                    month = 1
                if datetime.has_day():
                    day = datetime.get_day()
                else:
                    day = 1
            if exists_date and date.valid():
                year = date.get_year()
                month = date.get_month()
                day = date.get_day()

            if year is not None:
                gst_datetime = Gst.DateTime.new_local_time(
                    year, month, day, 0, 0, 0)
                glib_datetime = gst_datetime.to_g_date_time()
                timestamp = glib_datetime.to_unix()
            return (year, timestamp)
        except Exception as e:
            Logger.error("TagReader::get_year(): %s", e)
        return (None, None)

    def get_original_year(self, tags):
        """
            Return original release year
            @param tags as Gst.TagList
            @return year and timestamp (int, int)
        """
        def get_id3():
            date_string = self.__get_private_string(tags, "TDOR", False)
            try:
                date = get_iso_date_from_string(date_string)
                datetime = GLib.DateTime.new_from_iso8601(date, None)
                return (datetime.get_year(), datetime.to_unix())
            except:
                pass
            return (None, None)

        def get_ogg():
            try:
                date_string = self.__get_extended(tags, ['ORIGINALDATE'])
                date = get_iso_date_from_string(date_string)
                datetime = GLib.DateTime.new_from_iso8601(date, None)
                return (datetime.get_year(), datetime.to_unix())
            except:
                pass
            return (None, None)

        if tags is None:
            return None
        values = get_id3()
        if values[0] is None:
            values = get_ogg()
        return values

    def get_bpm(self, tags):
        """
            Get BPM from tags
            @param tags as Gst.TagList
            @return int/None
        """
        bpm = self.__get(tags, ["beats-per-minute"])
        if not bpm:
            bpm = None
        return bpm

    def get_popm(self, tags):
        """
            Get popularity tag
            @param tags as Gst.TagList
            @return int
        """
        try:
            if tags is None:
                return 0
            size = tags.get_tag_size("private-id3v2-frame")
            for i in range(0, size):
                (exists, sample) = tags.get_sample_index("private-id3v2-frame",
                                                         i)
                if not exists:
                    continue
                (exists, m) = sample.get_buffer().map(Gst.MapFlags.READ)
                if not exists:
                    continue
                # Gstreamer 1.18 API breakage
                try:
                    bytes = m.data.tobytes()
                except:
                    bytes = m.data

                if len(bytes) > 4 and bytes[0:4] == b"POPM":
                    try:
                        popm = bytes.split(b"\x00")[6][0]
                    except:
                        popm = 0
                    if popm == 0:
                        value = 0
                    elif popm >= 1 and popm < 64:
                        value = 1
                    elif popm >= 64 and popm < 128:
                        value = 2
                    elif popm >= 128 and popm < 196:
                        value = 3
                    elif popm >= 196 and popm < 255:
                        value = 4
                    elif popm == 255:
                        value = 5
                    else:
                        value = 0
                    return value
        except Exception as e:
            Logger.warning("TagReader::get_popm(): %s", e)
        return 0

    def get_lyrics(self, tags):
        """
            Return lyrics for tags
            @parma tags as Gst.TagList
            @return lyrics as str
        """
        def get_mp4():
            return self.__get(tags, ["lyrics"])

        def get_id3():
            return self.__get_private_string(tags, "USLT", True)

        def get_ogg():
            return self.__get_extended(tags, ["LYRICS"])

        if tags is None:
            return ""
        lyrics = get_mp4()
        if not lyrics:
            lyrics = get_id3()
        if not lyrics:
            lyrics = get_ogg()
        return lyrics

    def get_synced_lyrics(self, tags):
        """
            Return synced lyrics for tags
            @parma tags as Gst.TagList
            @return lyrics as ([str, int])
        """
        def decode_lyrics(bytes_list, encoding):
            lyrics = []
            try:
                for frame in bytes_list:
                    (l, t) = splitUnicode(frame, encoding)
                    if l:
                        lyrics.append((decodeUnicode(l, encoding),
                                       int.from_bytes(t[1:4], "big")))
            except Exception as e:
                Logger.warning(
                        "TagReader::get_synced_lyrics.decode_lyrics(): %s", e)
            return lyrics

        def get_id3():
            try:
                b = self.__get_private_bytes(tags, "SYLT")
                if b:
                    frame = b[10:]
                    encoding = frame[0:1]
                    string = decode_lyrics(frame.split(b"\n"), encoding)
                    if string is not None:
                        return string
            except Exception as e:
                Logger.warning("TagReader::get_synced_lyrics.get_id3(): %s", e)
            return ""

        if tags is None:
            return ""
        lyrics = get_id3()
        return lyrics

    def add_artists(self, artists, sortnames, mb_artist_id=""):
        """
            Add artists to db
            @param artists as str
            @param sortnames as str
            @param mb_artist_id as str
            @return ([int], [int]): (added artist ids, artist ids)
        """
        artist_ids = []
        added_artist_ids = []
        artistsplit = artists.split(";")
        sortsplit = sortnames.split(";")
        sortlen = len(sortsplit)
        mbidsplit = mb_artist_id.split(";")
        mbidlen = len(mbidsplit)
        if len(artistsplit) != mbidlen:
            mbidsplit = []
            mbidlen = 0
        i = 0
        for artist in artistsplit:
            artist = artist.strip()
            if artist != "":
                if i >= mbidlen or mbidsplit[i] == "":
                    mbid = None
                else:
                    mbid = mbidsplit[i].strip()
                # Get artist id, add it if missing
                (artist_id, db_name) = App().artists.get_id(artist, mbid)
                if i >= sortlen or sortsplit[i] == "":
                    sortname = None
                else:
                    sortname = sortsplit[i].strip()
                if artist_id is None:
                    if sortname is None:
                        sortname = format_artist_name(artist)
                    artist_id = App().artists.add(artist, sortname, mbid)
                    added_artist_ids.append(artist_id)
                else:
                    # artists.get_id() is NOCASE, check if we need to update
                    # artist name
                    if db_name != artist:
                        App().artists.set_name(artist_id, artist)
                    if sortname is not None:
                        App().artists.set_sortname(artist_id, sortname)
                    if mbid is not None:
                        App().artists.set_mb_artist_id(artist_id, mbid)
                i += 1
                artist_ids.append(artist_id)
        return (added_artist_ids, artist_ids)

    def add_genres(self, genres):
        """
            Add genres to db
            @param genres as string
            @return ([int], [int]): (added genre ids, genre ids)
        """
        genre_ids = []
        added_genre_ids = []
        for genre in genres.split(";"):
            genre = genre.strip()
            if genre != "":
                # Get genre id, add genre if missing
                genre_id = App().genres.get_id(genre)
                if genre_id is None:
                    genre_id = App().genres.add(genre)
                    added_genre_ids.append(genre_id)
                genre_ids.append(genre_id)
        return (added_genre_ids, genre_ids)

    def add_album(self, album_name, mb_album_id, lp_album_id, artist_ids,
                  uri, loved, popularity, rate, synced, mtime, storage_type):
        """
            Add album to db
            @param album_name as str
            @param mb_album_id as str
            @param lp_album_id as str
            @param artist_ids as [int]
            @param uri as str
            @param loved as bool
            @param popularity as int
            @param rate as int
            @param synced as int
            @param mtime as int
            @param storage_type as StorageType
            @return (added as bool, album_id as int)
            @commit needed
        """
        added = False
        if uri.find("://") != -1:
            f = Gio.File.new_for_uri(uri)
            parent = f.get_parent()
            if parent is not None:
                uri = parent.get_uri()
        album_id = App().albums.get_id(album_name, mb_album_id, artist_ids)
        # Check storage type did not changed, remove album then
        if album_id is not None:
            current_storage_type = App().albums.get_storage_type(album_id)
            if current_storage_type != storage_type:
                App().tracks.remove_album(album_id)
                App().tracks.clean(False)
                App().albums.clean(False)
                App().artists.clean(False)
                album_id = None
        if album_id is None:
            added = True
            album_id = App().albums.add(album_name, mb_album_id, lp_album_id,
                                        artist_ids, uri, loved, popularity,
                                        rate, synced, mtime, storage_type)
        # Check if path did not change
        elif App().albums.get_uri(album_id) != uri:
            App().albums.set_uri(album_id, uri)
        return (added, album_id)

#######################
# PRIVATE             #
#######################
    def __get_extended(self, tags, keys):
        """
            Return tag from tags following keys
            @param tags as Gst.TagList
            @param keys as [str]
            @return Tag as str
        """
        if tags is None:
            return ""
        items = []
        try:
            for i in range(tags.get_tag_size("extended-comment")):
                (exists, read) = tags.get_string_index("extended-comment", i)
                for key in keys:
                    if exists and read.startswith(key + "="):
                        items.append("".join(read.split("=")[1:]))
        except Exception as e:
            Logger.error("TagReader::__get_extended(): %s", e)
        return ";".join(items)

    def __get(self, tags, keys):
        """
            Return tag from tags following keys
            Only handles string/uint/double
            @param tags as Gst.TagList
            @param keys as [str]
            @return Tag as str/int/double. Empty string if does not exist
        """
        if tags is None:
            return ""
        items = []
        try:
            for key in keys:
                for i in range(tags.get_tag_size(key)):
                    if key in self.__STRING:
                        (exists, read) = tags.get_string_index(key, i)
                        if exists and read.strip(" "):
                            items.append(read)
                    elif key in self.__INT:
                        (exists, read) = tags.get_uint_index(key, i)
                        if exists:
                            return read
                    elif key in self.__DOUBLE:
                        (exists, read) = tags.get_double_index(key, i)
                        if exists:
                            return read
                    else:
                        Logger.error("Missing key" % key)
        except Exception as e:
            Logger.error("TagReader::__get(): %s", e)
        return ";".join(items)

    def __get_private_bytes(self, tags, key):
        """
            Get key from private frame
            @param tags as Gst.TagList
            @param key as str
            @return frame as bytes
        """
        try:
            size = tags.get_tag_size("private-id3v2-frame")
            encoded_key = key.encode("utf-8")
            for i in range(0, size):
                (exists, sample) = tags.get_sample_index(
                    "private-id3v2-frame",
                    i)
                if not exists:
                    continue
                (exists, m) = sample.get_buffer().map(Gst.MapFlags.READ)
                if not exists:
                    continue
                # Gstreamer 1.18 API breakage
                try:
                    b = m.data.tobytes()
                except:
                    b = m.data

                if b[0:len(encoded_key)] != encoded_key:
                    continue
                return b
        except Exception as e:
            Logger.error("TagReader::__get_private_bytes(): %s" % e)
        return b""

    def __get_private_string(self, tags, key, lang):
        """
            Get key from private frame
            @param tags as Gst.TagList
            @param key as str
            @param lang as bool
            @return Tag as str
        """
        try:
            b = self.__get_private_bytes(tags, key)
            if lang:
                frame = FrameLangTag(b)
            else:
                frame = FrameTextTag(b)
            if frame.key == key:
                return frame.string
        except Exception as e:
            Logger.error("TagReader::__get_private(): %s" % e)
        return ""



# Copyright (C) 2017 Jason Gray <jasonlevigray3@gmail.com>
# Copyright (C) 2017 Franz Dietrich <dietrich@teilgedanken.de>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

from gi.repository import Gtk, Gst

from gettext import gettext as _

from lollypop.define import App, PowerManagement


class Inhibitor:

    def __init__(self):
        self.__cookie = 0
        self.__status_handler_id = None      # The playback listener
        self.__current_player_state = None
        self.__manual_inhibit = False

        # Load and apply the inhibit settings
        self.__on_powermanagement_setting_changed(App().settings)
        # Register to settings changes
        App().settings.connect(
            "changed::power-management",
            self.__on_powermanagement_setting_changed,
        )

    def __on_powermanagement_setting_changed(self, settings, name=None):
        """
            Register to playback status changes so that standby/idle is only
            inhibited while playing
        """
        if settings.get_enum("power-management") > 0:
            self.__enable_react_to_playback()
        else:
            self.__disable_react_to_playback()
            self.__uninhibit()
        # Update the flags according to the settings.
        self.__update_flags_settings()

    def __disable_react_to_playback(self):
        if self.__status_handler_id is not None:
            App().player.disconnect(self.__status_handler_id)

    def __enable_react_to_playback(self):
        self.__on_status_changed(App().player)
        if self.__status_handler_id is None:
            self.__status_handler_id = App().player.connect(
                "status-changed",
                self.__on_status_changed,
            )

    def __on_status_changed(self, player):
        """
            React to a change of playback state
        """
        new_state = player.get_status()
        if self.__current_player_state != new_state:
            self.__current_player_state = new_state
            if self.__current_player_state == Gst.State.PLAYING:
                self.__update_flags_settings()
            else:
                self.__uninhibit()

    def __update_flags_settings(self):
        """
            Update the inhibit flags according to the settings in dconf
        """
        power_management = App().settings.get_enum("power-management")
        if power_management == PowerManagement.BOTH:
            self.__inhibit(Gtk.ApplicationInhibitFlags.IDLE |
                           Gtk.ApplicationInhibitFlags.SUSPEND)
        elif power_management == PowerManagement.SUSPEND:
            self.__inhibit(Gtk.ApplicationInhibitFlags.SUSPEND)
        elif power_management == PowerManagement.IDLE:
            self.__inhibit(Gtk.ApplicationInhibitFlags.IDLE)

    def __inhibit(self, flags):
        """
            Disable flags
            @param flags as Gtk.ApplicationInhibitFlags
        """
        if self.__manual_inhibit:
            # temporary blocked inhibit changes
            return
        if not self.__cookie:
            self.__cookie = App().inhibit(
                App().window,
                flags,
                _("Playing music"))

    def __uninhibit(self):
        """
            Remove all the powermanagement settings
        """
        if self.__manual_inhibit:
            # temporary blocked inhibit changes
            return
        if self.__cookie:
            App().uninhibit(self.__cookie)
            self.__cookie = 0
        self.__current_player_state = None

    def manual_inhibit(self, flags):
        """
            Inhibit suspend or idle manually.
            The settings values from dconf are not applied while a
            manual_inhibit() call is active.
            Disable the manual override with manual_uninhibit().
            @param flags as Gtk.ApplicationInhibitFlags
        """
        self.__uninhibit()
        self.__inhibit(flags)
        self.__manual_inhibit = True

    def manual_uninhibit(self):
        """
            removing the manual inhibited state and restore the settings from
            dconf
        """
        self.__manual_inhibit = False
        self.__uninhibit()
        self.__update_flags_settings()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject

from lollypop.define import App
from lollypop.utils import emit_signal
from lollypop.helper_signals import SignalsHelper, signals_map


class ApplicationMenu(Gtk.Bin, SignalsHelper):
    """
        Configure defaults items
    """

    __gsignals__ = {
        "hidden": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    @signals_map
    def __init__(self):
        """
            Init popover
        """
        Gtk.Bin.__init__(self)
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/Appmenu.ui")
        widget = builder.get_object("widget")
        self.add(widget)
        self.__volume = builder.get_object("volume")
        self.__volume.set_value(App().player.volume)
        builder.connect_signals(self)
        if App().settings.get_value("background-mode"):
            builder.get_object("quit_button").show()
        if App().window.folded:
            builder.get_object("shortcuts_button").hide()
        return [
                (App().player, "volume-changed", "_on_volume_changed")
        ]

#######################
# PROTECTED           #
#######################
    def _on_button_clicked(self, button):
        """
            Popdown popover if exists
            @param button as Gtk.Button
        """
        popover = self.get_ancestor(Gtk.Popover)
        if popover is not None:
            popover.popdown()
        else:
            emit_signal(self, "hidden", True)

    def _emit_hidden(self, button):
        """
            Emit hidden signal
        """
        emit_signal(self, "hidden", False)

    def _on_volume_value_changed(self, scale):
        """
            Set volume
            @param scale as Gtk.Scale
        """
        new_volume = scale.get_value()
        if new_volume != App().player.volume:
            App().player.set_volume(scale.get_value())

    def _on_volume_changed(self, player):
        """
            Set scale value
            @param player as Player
        """
        volume = self.__volume.get_value()
        if player.volume != volume:
            self.__volume.set_value(player.volume)

    def _mute_volume(self, event_box, event_button):
        """
            Mute the volume
            @param event_box as Gtk.EventBox
            @param event_button as Gdk.EventButton
        """
        self.__volume.set_value(0)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject, Gtk, Gdk

from lollypop.define import App
from lollypop.utils import emit_signal
from lollypop.helper_signals import SignalsHelper, signals_map
from lollypop.helper_gestures import GesturesHelper


class TracksWidget(Gtk.ListBox, SignalsHelper, GesturesHelper):
    """
        Widget showing a track list
    """

    __gsignals__ = {
        "activated": (GObject.SignalFlags.RUN_FIRST,
                      None, (GObject.TYPE_PYOBJECT,)),
        "do-selection": (GObject.SignalFlags.RUN_FIRST,
                         None, (GObject.TYPE_PYOBJECT,)),
        "do-shift-selection": (GObject.SignalFlags.RUN_FIRST,
                               None, (GObject.TYPE_PYOBJECT,))
    }

    @signals_map
    def __init__(self, view_type):
        """
            Init track widget
            @param view_type as ViewType
        """
        Gtk.ListBox.__init__(self)
        GesturesHelper.__init__(self, self)
        self.__view_type = view_type
        self.get_style_context().add_class("trackswidget")
        self.set_property("hexpand", True)
        self.set_selection_mode(Gtk.SelectionMode.NONE)
        return [
            (App().player, "queue-changed", "_on_queue_changed")
        ]

    def update_playing(self, track_id):
        """
            Update playing track
            @param track_id as int
        """
        for row in self.get_children():
            row.set_indicator()

    def update_duration(self, track_id):
        """
            Update duration for track id
            @param track_id as int
        """
        for row in self.get_children():
            if row.track.id == track_id:
                row.update_duration()

#######################
# PROTECTED           #
#######################
    def _on_queue_changed(self, *ignore):
        """
            Update all position labels
        """
        for row in self.get_children():
            row.update_number_label()

    def _on_primary_long_press_gesture(self, x, y):
        """
            Show row menu
            @param x as int
            @param y as int
        """
        row = self.get_row_at_y(y)
        if row is None:
            return
        emit_signal(self, "do-selection", row)

    def _on_primary_press_gesture(self, x, y, event):
        """
            Activate current row
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        row = self.get_row_at_y(y)
        if row is None:
            return
        if event.state & Gdk.ModifierType.SHIFT_MASK:
            emit_signal(self, "do-shift-selection", row)
        elif event.state & Gdk.ModifierType.CONTROL_MASK:
            emit_signal(self, "do-selection", row)
        elif event.state & Gdk.ModifierType.MOD1_MASK:
            emit_signal(self, "do-selection", None)
            App().player.clear_albums()
            App().player.load(row.track)
        else:
            emit_signal(self, "activated", row.track)
            emit_signal(self, "do-selection", None)

    def _on_secondary_press_gesture(self, x, y, event):
        """
            Show row menu
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        row = self.get_row_at_y(y)
        if row is None:
            return
        row.popup_menu(self, x, y)

#######################
# PRIVATE             #
#######################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gio

from re import sub

from lollypop.helper_web_base import BaseWebHelper
from lollypop.utils import emit_signal
from lollypop.utils_file import get_youtube_dl
from lollypop.logger import Logger


class YouTubeWebHelper(BaseWebHelper):
    """
        YoutTube helper
    """

    __BAD_SCORE = 1000000

    def __init__(self):
        """
            Init heApper
        """
        BaseWebHelper.__init__(self)

    def get_uri_content(self, uri, cancellable):
        """
            Get content uri
            @param uri as str
            @param cancellable as Gio.Cancellable
            @return content uri as str/None
        """
        Logger.info("Loading %s with YouTube", uri)
        try:
            proxy = GLib.environ_getenv(GLib.get_environ(), "all_proxy")
            if proxy is not None and proxy.startswith("socks://"):
                proxy = proxy.replace("socks://", "socks4://")
            (path, env) = get_youtube_dl()
            # Remove playlist args
            uri = sub("list=.*", "", uri)
            argv = [path, "--no-cache-dir", "-g", "-f", "bestaudio", uri]
            if proxy is not None:
                argv += ["--proxy", proxy, None]
            else:
                argv.append(None)
            process = Gio.Subprocess.new(argv, Gio.SubprocessFlags.STDOUT_PIPE)
            process.wait_async(cancellable, self.__on_youtube_dl, cancellable)
        except Exception as e:
            Logger.error("YouTubeWebHelper::get_uri_content(): %s", e)

#######################
# PRIVATE             #
#######################
    def __on_youtube_dl(self, process, result, cancellable):
        """
            Emit signal for content
            @param process as Gio.Subprocess.
            @param result as Gio.AsyncResult
            @param cancellable as Gio.Cancellable
        """
        try:
            content = ""
            status = process.wait_check_finish(result)
            if status:
                stream = process.get_stdout_pipe()
                bytes = bytearray(0)
                buf = stream.read_bytes(1024, cancellable).get_data()
                while buf:
                    bytes += buf
                    buf = stream.read_bytes(1024, cancellable).get_data()
                stream.close()
                content = bytes.decode("utf-8")
        except Exception as e:
            Logger.warning("YouTubeWebHelper::__on_youtube_dl(): %s", e)
        emit_signal(self, "uri-content-loaded", content)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk, Gtk, GLib, GObject

from lollypop.objects_album import Album
from lollypop.utils import set_cursor_type
from lollypop.widgets_row_album import AlbumRow
from lollypop.widgets_row_track import TrackRow


class DNDHelper(GObject.Object):
    """
        Helper for DND of AlbumsListView
    """

    __gsignals__ = {
        "dnd-insert": (GObject.SignalFlags.RUN_FIRST, None,
                       (GObject.TYPE_PYOBJECT, int)),
        "dnd-finished": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, listbox, view_type):
        """
            Init helper
            @param listbox as Gtk.ListBox
            @params view_type as ViewType
        """
        GObject.Object.__init__(self)
        self.__listbox = listbox
        self.__view_type = view_type
        self.__drag_begin_rows = []
        self.__autoscroll_timeout_id = None
        self.__begin_scrolled_y = 0
        self.__gesture = Gtk.GestureDrag.new(listbox)
        self.__gesture.set_button(Gdk.BUTTON_PRIMARY)
        self.__gesture.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        self.__gesture.connect("drag-begin", self.__on_drag_begin)
        self.__gesture.connect("drag-end", self.__on_drag_end)
        self.__gesture.connect("drag-update", self.__on_drag_update)

    @property
    def gesture(self):
        """
            Get gesture
            @return Gtk.GestureDrag
        """
        return self.__gesture

#######################
# PRIVATE             #
#######################
    def __update_album_rows(self, start_index, end_index):
        """
            Merge album rows and update track position for indexes
        """
        def merge_if_possible(row1, row2):
            if row1 is None or row2 is None:
                return False
            if row1.album.id == row2.album.id:
                row1.tracks_view.append_rows(row2.album.tracks)
                row1.reveal(True)
                return True
            return False

        children = self.__listbox.get_children()[start_index:end_index]
        while children:
            row = children[0]
            index = row.get_index()
            next = self.__listbox.get_row_at_index(index + 1)
            if merge_if_possible(row, next):
                if next in children:
                    children.remove(next)
                self.__listbox.remove(next)
            else:
                children.pop(0)
        GLib.timeout_add(100, self.emit, "dnd-finished")

    def __do_drag_and_drop(self, src_rows, dest_row, direction):
        """
            Drag source rows at destination row with direction
            @param src_rows as [Row]
            @param dest_row as Row
            @param direction as Gtk.DirectionType
        """
        indexes = []
        # Build new rows
        new_rows = self.__get_rows_from_rows(
            src_rows, AlbumRow.get_best_height(dest_row))
        # Insert new rows
        if isinstance(dest_row, TrackRow):
            album_row = dest_row.get_ancestor(AlbumRow)
            if album_row is None:
                return
            # Destroy src_rows from album before split
            for row in src_rows:
                if isinstance(row, TrackRow):
                    if row.get_ancestor(AlbumRow) == album_row:
                        album_row.tracks_view.remove_row(row.track)
            indexes.append(album_row.get_index())
            split_album_row = self.__split_album_row(album_row,
                                                     dest_row,
                                                     direction)
            index = album_row.get_index()
            if split_album_row is not None:
                self.emit("dnd-insert", split_album_row, index)
                index += 1
            elif direction == Gtk.DirectionType.DOWN:
                index += 1
            for row in new_rows:
                self.emit("dnd-insert", row, index)
                index += 1
        else:
            index = dest_row.get_index()
            indexes.append(index)
            if direction == Gtk.DirectionType.DOWN:
                index += 1
            for row in new_rows:
                self.emit("dnd-insert", row, index)
                index += 1
        # Calculate update range
        for row in src_rows:
            if isinstance(row, TrackRow):
                album_row = row.get_ancestor(AlbumRow)
                if album_row is not None:
                    indexes.append(album_row.get_index())
            else:
                indexes.append(row.get_index())
        self.__destroy_rows(src_rows)
        self.__update_album_rows(max(0, min(indexes) - 1), max(indexes) + 1)

    def __split_album_row(self, album_row, track_row, direction):
        """
            Split album row at track row with direction
            @param album_row as AlbumRow
            @param track_row as TrackRow
            @param direction as Gtk.DirectionType
            @return AlbumRow
        """
        height = AlbumRow.get_best_height(album_row)
        children = album_row.children
        index = children.index(track_row)
        if direction == Gtk.DirectionType.DOWN:
            index += 1
            if index + 1 > len(children):
                return None
        elif index - 1 < 0:
            return None
        rows = album_row.children[:index]
        split_album = Album(album_row.album.id)
        split_album.set_tracks([row.track for row in rows])
        split_album_row = AlbumRow(split_album, height, self.__view_type)
        split_album_row.reveal()
        split_album_row.show()
        for row in rows:
            empty = album_row.album.remove_track(row.track)
            if empty:
                album_row.destroy()
            row.destroy()
        return split_album_row

    def __get_rows_from_rows(self, rows, height):
        """
            Build news rows from rows
            @param rows as [TrackRow/AlbumRow]
            @param height as int
            @return [AlbumRow]
        """
        new_rows = []
        for row in rows:
            if isinstance(row, TrackRow):
                # Merge with previous
                if new_rows and new_rows[-1].album.id == row.track.album.id:
                    new_rows[-1].tracks_view.append_row(row.track)
                # Create a new album
                else:
                    new_album = Album(row.track.album.id)
                    new_album.set_tracks([row.track])
                    new_album_row = AlbumRow(new_album, height,
                                             self.__view_type)
                    new_album_row.show()
                    new_album_row.reveal()
                    new_rows.append(new_album_row)
            else:
                # Merge with previous
                if new_rows and new_rows[-1].album.id == row.album.id:
                    new_rows[-1].tracks_view.append_rows(row.album.tracks)
                # Create a new album
                else:
                    new_album = Album(row.album.id)
                    new_album.set_tracks(row.album.tracks)
                    new_album_row = AlbumRow(new_album, height,
                                             self.__view_type)
                    new_album_row.populate()
                    new_album_row.show()
                    new_rows.append(new_album_row)
        return new_rows

    def __destroy_rows(self, rows):
        """
            Destroy rows and parent if needed
            @param rows as [Row]
        """
        for row in rows:
            if isinstance(row, TrackRow):
                album_row = row.get_ancestor(AlbumRow)
                if album_row is not None:
                    album_row.album.remove_track(row.track)
                    if album_row.album.id is None:
                        album_row.destroy()
            row.destroy()

    def __unmark_all_rows(self):
        """
            Undrag all rows
        """
        for row in self.__listbox.get_children():
            context = row.get_style_context()
            context.remove_class("drag-up")
            context.remove_class("drag-down")
            if row.revealed:
                for subrow in row.children:
                    context = subrow.get_style_context()
                    context.remove_class("drag-up")
                    context.remove_class("drag-down")

    def __get_row_at_y(self, y):
        """
            Get row at position
            @param y as int
            @return (Gtk.ListBox, Row)
        """
        row = self.__listbox.get_row_at_y(y)
        if row is not None and row.revealed:
            (listbox, subrow) = self.__get_subrow_at_y(row, y)
            if subrow is not None:
                return (listbox, subrow)
        return (self.__listbox, row)

    def __get_subrow_at_y(self, album_row, y):
        """
            Get subrow as position
            @param album_row as AlbumRow
            @param y as int
            @return (Gtk.ListBox, Row)
        """
        if album_row is not None:
            listbox = album_row.listbox
            t = listbox.translate_coordinates(self.__listbox, 0, 0)
            if t is not None:
                track_row = listbox.get_row_at_y(y - t[1])
                if track_row is not None:
                    return (listbox, track_row)
        return (None, None)

    def __autoscroll(self, scrolled, y):
        """
            Auto scroll up/down
            @param scrolled as Gtk.ScrolledWindow
            @param y as int
        """
        adj = scrolled.get_vadjustment()
        value = adj.get_value()
        adj_value = value + y
        adj.set_value(adj_value)
        if adj.get_value() <= adj.get_lower() or\
                adj.get_value() >= adj.get_upper() - adj.get_page_size():
            self.__autoscroll_timeout_id = None
            return False
        return True

    def __on_drag_begin(self, gesture, x, y):
        """
            @param gesture as Gtk.GestureDrag
            @param x as int
            @param y as int
        """
        self.__drag_begin_rows = []
        (listbox, row) = self.__get_row_at_y(y)
        if row is not None:
            self.__drag_begin_rows += [row]
        for row in listbox.get_selected_rows():
            if row not in self.__drag_begin_rows:
                self.__drag_begin_rows.append(row)
        scrolled = self.__listbox.get_ancestor(Gtk.ScrolledWindow)
        if scrolled is not None and row is not None:
            (scrolled_x,
             self.__begin_scrolled_y) = row.translate_coordinates(scrolled,
                                                                  0, 0)

    def __on_drag_end(self, gesture, x, y):
        """
            @param gesture as Gtk.GestureDrag
            @param x as int
            @param y as int
        """
        set_cursor_type(self.__listbox, "default")
        self.__unmark_all_rows()
        if self.__autoscroll_timeout_id is not None:
            GLib.source_remove(self.__autoscroll_timeout_id)
            self.__autoscroll_timeout_id = None
        if x == 0 or y == 0:
            return
        (active, start_x, start_y) = gesture.get_start_point()
        if not active:
            return
        y += start_y
        (listbox, row) = self.__get_row_at_y(y)
        if row is None or row in self.__drag_begin_rows:
            return
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        if self.__drag_begin_rows:
            row_height = row.get_allocated_height()
            (row_x, row_y) = row.translate_coordinates(self.__listbox,
                                                       0, 0)
            if y < row_y + row_height / 2:
                direction = Gtk.DirectionType.UP
            elif y > row_y - row_height / 2:
                direction = Gtk.DirectionType.DOWN
            GLib.idle_add(self.__do_drag_and_drop,
                          self.__drag_begin_rows,
                          row, direction)

    def __on_drag_update(self, gesture, x, y):
        """
            Add style
            @param gesture as Gtk.GestureDrag
            @param x as int
            @param y as int
        """
        if self.__autoscroll_timeout_id is not None:
            GLib.source_remove(self.__autoscroll_timeout_id)
            self.__autoscroll_timeout_id = None
        self.__unmark_all_rows()
        set_cursor_type(self.__listbox, "dnd-move")
        (active, start_x, start_y) = gesture.get_start_point()
        if not active:
            return
        current_y = y + start_y
        (ignore, row) = self.__get_row_at_y(current_y)
        if row is None:
            return
        row_height = row.get_allocated_height()
        (row_x, row_y) = row.translate_coordinates(self.__listbox, 0, 0)
        if current_y < row_y + 20:
            row.get_style_context().add_class("drag-up")
        elif current_y > row_y + row_height - 20:
            row.get_style_context().add_class("drag-down")
        scrolled = self.__listbox.get_ancestor(Gtk.ScrolledWindow)
        if scrolled is None:
            return
        (scrolled_x, scrolled_y) = row.translate_coordinates(scrolled, 0, 0)
        diff = self.__begin_scrolled_y - scrolled_y
        if abs(diff) < 100:
            return
        self.__autoscroll_timeout_id = GLib.idle_add(self.__autoscroll,
                                                     scrolled,
                                                     -diff / 10000)
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Pango, GLib, GObject

from gettext import gettext as _

from lollypop.define import App, ViewType, MARGIN_SMALL, IndicatorType
from lollypop.define import StorageType, LovedFlags
from lollypop.widgets_indicator import IndicatorWidget
from lollypop.utils import ms_to_string, on_query_tooltip, popup_widget
from lollypop.utils import emit_signal


class TrackRow(Gtk.ListBoxRow):
    """
        A track row
    """

    __gsignals__ = {
        "removed": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def get_best_height(widget):
        """
            Calculate widget height
            @param widget as Gtk.Widget
        """
        ctx = widget.get_pango_context()
        layout = Pango.Layout.new(ctx)
        layout.set_text("a", 1)
        font_height = int(layout.get_pixel_size()[1])
        # application.css
        min_height = 32
        if font_height > min_height:
            height = font_height
        else:
            height = min_height
        return height

    def __init__(self, track, album_artist_ids, view_type):
        """
            Init row widgets
            @param track as Track
            @param album_artist_ids as [int]
            @param view_type as ViewType
        """
        # We do not use Gtk.Builder for speed reasons
        Gtk.ListBoxRow.__init__(self)
        self.__view_type = view_type
        self._track = track
        self.set_property("margin", MARGIN_SMALL)
        self._grid = Gtk.Grid()
        self._grid.set_property("valign", Gtk.Align.CENTER)
        self._grid.set_column_spacing(5)
        self._grid.show()
        if not view_type & ViewType.QUEUE:
            self._indicator = IndicatorWidget(self, view_type)
            self._indicator.show()
            self._indicator.set_valign(Gtk.Align.CENTER)
            self._grid.add(self._indicator)
        self._num_label = Gtk.Label.new()
        self._num_label.set_ellipsize(Pango.EllipsizeMode.END)
        self._num_label.set_width_chars(4)
        self._num_label.get_style_context().add_class("dim-label")
        self.update_number_label()
        self._grid.add(self._num_label)
        self.__title_label = Gtk.Label.new()
        self.__title_label.set_property("has-tooltip", True)
        self.__title_label.connect("query-tooltip", on_query_tooltip)
        self.__title_label.set_property("hexpand", True)
        self.__title_label.set_property("halign", Gtk.Align.START)
        self.__title_label.set_property("xalign", 0)
        self.__title_label.set_ellipsize(Pango.EllipsizeMode.END)
        self.__title_label.show()
        self._grid.add(self.__title_label)
        featuring_artist_ids = track.get_featuring_artist_ids(album_artist_ids)
        markup = GLib.markup_escape_text(self._track.name)
        if featuring_artist_ids:
            artists = []
            for artist_id in featuring_artist_ids:
                artists.append(App().artists.get_name(artist_id))
            markup += "\n<span size='x-small' alpha='40000'>%s</span>" %\
                GLib.markup_escape_text(", ".join(artists))
        self.__title_label.set_markup(markup)
        duration = ms_to_string(self._track.duration)
        self.__duration_label = Gtk.Label.new(duration)
        self.__duration_label.get_style_context().add_class("dim-label")
        self.__duration_label.show()
        self._grid.add(self.__duration_label)
        self.__action_button = None
        if self.__view_type & (
                ViewType.PLAYBACK | ViewType.PLAYLISTS):
            self.__action_button = Gtk.Button.new_from_icon_name(
               "list-remove-symbolic",
               Gtk.IconSize.MENU)
            self.__action_button.set_tooltip_text(
               _("Remove from playlist"))
        elif self.__view_type & (ViewType.ALBUM | ViewType.ARTIST):
            self.__action_button = Gtk.Button.new_from_icon_name(
               "view-more-symbolic",
               Gtk.IconSize.MENU)
        if self.__action_button is None:
            self.__duration_label.set_margin_end(MARGIN_SMALL)
        else:
            self.__action_button.show()
            self.__action_button.connect("clicked",
                                         self.__on_action_button_clicked)
            self.__action_button.set_margin_end(MARGIN_SMALL)
            self.__action_button.set_relief(Gtk.ReliefStyle.NONE)
            context = self.__action_button.get_style_context()
            context.add_class("menu-button")
            self._grid.add(self.__action_button)
        self.add(self._grid)
        if not view_type & ViewType.QUEUE:
            self.set_indicator(self._get_indicator_type())
        self.update_duration()
        self.get_style_context().add_class("trackrow")

    def update_duration(self):
        """
            Update track duration
        """
        self._track.reset("duration")
        duration = ms_to_string(self._track.duration)
        self.__duration_label.set_label(duration)

    def set_indicator(self, indicator_type=None):
        """
            Show indicator
            @param indicator_type as IndicatorType
        """
        if indicator_type is None:
            indicator_type = self._get_indicator_type()
        self._indicator.clear()
        if indicator_type & IndicatorType.LOADING:
            self._indicator.set_opacity(1)
            self._indicator.load()
        elif indicator_type & IndicatorType.PLAY:
            self._indicator.set_opacity(1)
            self.set_state_flags(Gtk.StateFlags.VISITED, True)
            if indicator_type & IndicatorType.LOVED:
                self._indicator.play_loved()
            else:
                self._indicator.play()
        else:
            self.unset_state_flags(Gtk.StateFlags.VISITED)
            if indicator_type & IndicatorType.LOVED:
                self._indicator.set_opacity(1)
                self._indicator.loved()
            elif indicator_type & IndicatorType.SKIPPED:
                self._indicator.set_opacity(1)
                self._indicator.skip()
            else:
                self._indicator.set_opacity(0)

    def update_number_label(self):
        """
            Update position label for row
        """
        if App().player.is_in_queue(self._track.id):
            self._num_label.get_style_context().add_class("queued")
            pos = App().player.get_track_position(self._track.id)
            self._num_label.set_text(str(pos))
            self._num_label.show()
        elif not self.__view_type & ViewType.SEARCH:
            if self.__view_type & (ViewType.PLAYBACK | ViewType.PLAYLISTS):
                discs = App().albums.get_discs(self._track.album.id)
                if len(discs) > 1:
                    label = "%s: %s" % (self._track.discnumber,
                                        self._track.number)
                else:
                    label = "%s" % self._track.number
            else:
                label = "%s" % self._track.number
            self._num_label.set_markup(label)
            self._num_label.get_style_context().remove_class("queued")
            self._num_label.show()
        else:
            self._num_label.hide()

    def popup_menu(self, parent, x=None, y=None):
        """
            Popup menu for track
            @param parent as Gtk.Widget
            @param x as int
            @param y as int
        """
        def on_hidden(widget, hide):
            self.set_indicator()

        from lollypop.menu_objects import TrackMenu, TrackMenuExt
        from lollypop.widgets_menu import MenuBuilder
        menu = TrackMenu(self._track, self.__view_type)
        menu_widget = MenuBuilder(menu)
        menu_widget.show()
        if not self._track.storage_type & StorageType.EPHEMERAL:
            menu_ext = TrackMenuExt(self._track)
            menu_ext.show()
            menu_widget.add_widget(menu_ext)
        if self.is_selected():
            popover = popup_widget(menu_widget, parent, x, y, None)
        else:
            popover = popup_widget(menu_widget, parent, x, y, self)
        if popover is None:
            menu_widget.connect("hidden", on_hidden)
        else:
            popover.connect("hidden", on_hidden)

    @property
    def name(self):
        """
            Get row name
            @return str
        """
        return self.__title_label.get_text()

    @property
    def track(self):
        """
            Get row track
            @return Track
        """
        return self._track

#######################
# PROTECTED           #
#######################
    def _get_indicator_type(self):
        """
            Get indicator type for current row
            @return IndicatorType
        """
        indicator_type = IndicatorType.NONE
        if App().player.current_track.id == self._track.id:
            indicator_type |= IndicatorType.PLAY
        if self._track.loved & LovedFlags.LOVED:
            indicator_type |= IndicatorType.LOVED
        elif self._track.loved & LovedFlags.SKIPPED:
            indicator_type |= IndicatorType.SKIPPED
        return indicator_type

#######################
# PRIVATE             #
#######################
    def __on_action_button_clicked(self, button):
        """
            Show row menu
            @param button as Gtk.Button
        """
        if self.__view_type & (
                ViewType.PLAYBACK | ViewType.QUEUE | ViewType.PLAYLISTS):
            emit_signal(self, "removed")
        else:
            self.popup_menu(button)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Pango, GObject

from gettext import gettext as _

from lollypop.define import App, MARGIN, ViewType, ArtBehaviour
from lollypop.widgets_banner import BannerWidget
from lollypop.utils import emit_signal
from lollypop.helper_signals import SignalsHelper, signals_map


class LyricsBannerWidget(BannerWidget, SignalsHelper):
    """
        Banner for lyrics
    """

    __gsignals__ = {
        "translate": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    @signals_map
    def __init__(self, view_type=ViewType.OVERLAY):
        """
            Init cover widget
            @param album
            @param view_type as ViewType
        """
        BannerWidget.__init__(self, view_type)
        self.__title_label = Gtk.Label.new()
        self.__title_label.show()
        self.__title_label.set_ellipsize(Pango.EllipsizeMode.END)
        self.__title_label.set_vexpand(True)
        self.__translate_button = Gtk.ToggleButton.new()
        image = Gtk.Image.new_from_icon_name(
            "accessories-dictionary-symbolic", Gtk.IconSize.BUTTON)
        image.show()
        self.__translate_button.set_image(image)
        self.__translate_button.show()
        self.__translate_button.connect("toggled",
                                        self.__on_lyrics_button_toggled)
        self.__translate_button.get_style_context().add_class(
            "banner-button")
        self.__translate_button.set_property("valign", Gtk.Align.CENTER)
        self.__translate_button.set_property("halign", Gtk.Align.END)
        self.__translate_button.set_hexpand(True)
        self.__translate_button.set_sensitive(False)
        grid = Gtk.Grid()
        grid.show()
        grid.set_column_spacing(MARGIN)
        grid.add(self.__title_label)
        grid.add(self.__translate_button)
        grid.set_property("margin", MARGIN)
        grid.set_margin_end(MARGIN)
        self._overlay.add_overlay(grid)
        self._overlay.set_overlay_pass_through(grid, True)
        self.__update()
        self.__set_internal_size()
        return [
            (App().player, "current-changed", "_on_current_changed"),
            (App().window.container.widget, "notify::folded",
             "_on_container_folded"),
            (App().album_art, "album-artwork-changed",
             "_on_album_artwork_changed")
        ]

    def update_for_width(self, width):
        """
            Update banner internals for width, call this before showing banner
            @param width as int
        """
        BannerWidget.update_for_width(self, width)
        self.__set_artwork()

    @property
    def translate_button(self):
        """
            Get translate button
            @return Gtk.Button
        """
        return self.__translate_button

#######################
# PROTECTED           #
#######################
    def _handle_width_allocate(self, allocation):
        """
            Update artwork
            @param allocation as Gtk.Allocation
        """
        if BannerWidget._handle_width_allocate(self, allocation):
            self.__set_artwork()

    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        self.__set_internal_size()

    def _on_album_artwork_changed(self, art, album_id):
        """
            Update cover for album_id
            @param art as Art
            @param album_id as int
        """
        if album_id == App().player.current_track.album.id:
            self.__set_artwork()

    def _on_current_changed(self, player):
        """
            Update labels and artwork
            @param player as Player
        """
        self.__update()

#######################
# PRIVATE             #
#######################
    def __update(self):
        """
            Update banner
        """
        if App().player.current_track.id is None:
            self.__title_label.set_text(_("No track playing"))
            self._on_artwork(None)
        else:
            markup = "%s\n" % GLib.markup_escape_text(
                App().player.current_track.name)
            artist_name = GLib.markup_escape_text(
                ", ".join(App().player.current_track.artists))
            markup += "<span size='x-small' alpha='40000'>%s</span>" %\
                artist_name
            self.__title_label.set_markup(markup)
            self.__set_artwork()

    def __set_artwork(self):
        """
            Set artwork
        """
        if not App().animations:
            return
        if App().player.current_track.id is None:
            self._artwork.get_style_context().add_class(
                "default-banner")
        else:
            self._artwork.get_style_context().remove_class(
                "default-banner")
            App().art_helper.set_album_artwork(
                    App().player.current_track.album,
                    # +100 to prevent resize lag
                    self.width + 100,
                    self.height,
                    self._artwork.get_scale_factor(),
                    ArtBehaviour.BLUR_HARD |
                    ArtBehaviour.DARKER,
                    self._on_artwork)

    def __set_internal_size(self):
        """
            Set text height
        """
        title_context = self.__title_label.get_style_context()
        for c in title_context.list_classes():
            title_context.remove_class(c)
        if App().window.folded:
            self.__title_label.get_style_context().add_class(
                "text-large")
        else:
            self.__title_label.get_style_context().add_class(
                "text-x-large")

    def __on_lyrics_button_toggled(self, button):
        """
            Emit signals
            @param button as Gtk.ToggleButton
        """
        emit_signal(self, "translate", button.get_active())



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json

from lollypop.logger import Logger
from lollypop.utils import emit_signal, get_network_available
from lollypop.helper_web_jamendo import JamendoWebHelper
from lollypop.helper_web_save import SaveWebHelper
from lollypop.define import App


class JamendoSearch(SaveWebHelper, JamendoWebHelper):
    """
        Search for Jamendo
    """

    def __init__(self):
        """
            Init object
        """
        SaveWebHelper.__init__(self)
        JamendoWebHelper.__init__(self)

    def get(self, search, storage_type, cancellable):
        """
            Get albums for search
            We need a thread because we are going to populate DB
            @param search as str
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
        """
        if not get_network_available("DEEZER"):
            emit_signal(self, "finished")
            return
        try:
            # Albums
            uri = "https://api.jamendo.com/v3.0/albums/"
            uri += "?client_id=6a13ad2c&format=jsonpretty"
            uri += "&limit=10&namesearch=%s" % search
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                for album in decode["results"]:
                    payload = self.lollypop_album_payload(album)
                    self.save_album_payload_to_db(payload,
                                                  storage_type,
                                                  True,
                                                  cancellable)
            # Artists
            uri = "https://api.jamendo.com/v3.0/artists/albums/"
            uri += "?client_id=6a13ad2c&format=jsonpretty"
            uri += "&limit=10&namesearch=%s" % search
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                for artist in decode["results"]:
                    App().art.add_from_uri(artist["name"], artist["image"],
                                           cancellable, storage_type)
                    for album in artist["albums"]:
                        album["artist_name"] = artist["name"]
                        payload = self.lollypop_album_payload(album)
                        self.save_album_payload_to_db(payload,
                                                      storage_type,
                                                      True,
                                                      cancellable)
        except Exception as e:
            Logger.warning("JamendoSearch::get(): %s", e)
        if not cancellable.is_cancelled():
            emit_signal(self, "finished")

    def load_tracks(self, album, cancellable):
        """
            Load tracks for album
            @param album as Album
            @param cancellable as Gio.Cancellable
        """
        try:
            jamid = album.uri.replace("jam:", "")
            uri = "https://api.jamendo.com/v3.0/albums/tracks/"
            uri += "?client_id=6a13ad2c&format=jsonpretty&id=%s" % jamid
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                # We want to share the same item as lp_album_id may change
                album_item = album.collection_item
                for result in decode["results"]:
                    for track in result["tracks"]:
                        track["artist_name"] = result["artist_name"]
                        lollypop_payload = self.lollypop_track_payload(track)
                        self.save_track_payload_to_db(lollypop_payload,
                                                      album_item,
                                                      album.storage_type,
                                                      False,
                                                      cancellable)
        except Exception as e:
            Logger.error("JamendoSearch::load_tracks(): %s", e)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from gettext import gettext as _

from lollypop.define import App, ViewType, Type, LovedFlags
from lollypop.utils_album import tracks_to_albums
from lollypop.utils import get_default_storage_type, emit_signal
from lollypop.utils import get_network_available
from lollypop.objects_track import Track
from lollypop.objects_album import Album


class PlaybackMenu(Gio.Menu):
    """
        Base class for playback menu
    """

    def __init__(self):
        """
            Init menu
        """
        Gio.Menu.__init__(self)

    @property
    def in_player(self):
        """
            True if current object in player
            return bool
        """
        return False

#######################
# PROTECTED           #
#######################
    def _set_radio_action(self, artist_ids):
        """
            Set radio action
            @param artist_ids as [int]
        """
        radio_action = Gio.SimpleAction(name="radio_action_collection")
        App().add_action(radio_action)
        radio_action.connect("activate",
                             self.__on_radio_action_activate,
                             artist_ids)
        menu_item = Gio.MenuItem.new(_("Play a radio"),
                                     "app.radio_action_collection")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        self.append_item(menu_item)

    def _set_playback_actions(self):
        """
            Setup playback actions
        """
        if not self.in_player:
            append_playback_action = Gio.SimpleAction(
                name="append_playback_action")
            App().add_action(append_playback_action)
            append_playback_action.connect("activate",
                                           self._append_to_playback)
            menu_item = Gio.MenuItem.new(_("Add to playback"),
                                         "app.append_playback_action")
        else:
            del_playback_action = Gio.SimpleAction(name="del_playback_action")
            App().add_action(del_playback_action)
            del_playback_action.connect("activate",
                                        self._remove_from_playback)
            menu_item = Gio.MenuItem.new(_("Remove from playback"),
                                         "app.del_playback_action")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        self.append_item(menu_item)

#######################
# PRIVATE             #
#######################
    def __on_radio_action_activate(self, action, variant, artist_ids):
        """
            Play a radio from storage type
            @param Gio.SimpleAction
            @param GLib.Variant
            @param artist_ids as [int]
        """
        App().player.play_radio_from_collection(artist_ids)


class RadioPlaybackMenu(Gio.Menu):
    """
        Allow to play a radio from artist ids
    """

    def __init__(self, artist_ids):
        """
            Init menu
            @param artist_ids as [int]
        """
        Gio.Menu.__init__(self)
        section = Gio.Menu()
        radio_action = Gio.SimpleAction(name="radio_action_collection")
        App().add_action(radio_action)
        radio_action.connect("activate",
                             self.__on_radio_action_activate,
                             artist_ids)
        menu_item = Gio.MenuItem.new(_("Related tracks"),
                                     "app.radio_action_collection")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        section.append_item(menu_item)
        radio_action = Gio.SimpleAction(name="radio_action_loved")
        App().add_action(radio_action)
        radio_action.connect("activate",
                             self.__on_radio_action_activate,
                             artist_ids)
        menu_item = Gio.MenuItem.new(_("Loved tracks"),
                                     "app.radio_action_loved")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        section.append_item(menu_item)
        radio_action = Gio.SimpleAction(name="radio_action_populars")
        App().add_action(radio_action)
        radio_action.connect("activate",
                             self.__on_radio_action_activate,
                             artist_ids)
        menu_item = Gio.MenuItem.new(_("Popular tracks"),
                                     "app.radio_action_populars")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        section.append_item(menu_item)
        self.append_section(_("From collection"), section)
        section = Gio.Menu()
        radio_action = Gio.SimpleAction(name="radio_action_deezer")
        App().add_action(radio_action)
        radio_action.connect("activate",
                             self.__on_radio_action_activate,
                             artist_ids)
        radio_action.set_enabled(get_network_available("DEEZER"))
        menu_item = Gio.MenuItem.new(_("Deezer"),
                                     "app.radio_action_deezer")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        section.append_item(menu_item)
        radio_action = Gio.SimpleAction(name="radio_action_lastfm")
        App().add_action(radio_action)
        radio_action.connect("activate",
                             self.__on_radio_action_activate,
                             artist_ids)
        radio_action.set_enabled(get_network_available("LASTFM"))
        menu_item = Gio.MenuItem.new(_("Last.fm"),
                                     "app.radio_action_lastfm")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        section.append_item(menu_item)
        radio_action = Gio.SimpleAction(name="radio_action_spotify")
        App().add_action(radio_action)
        radio_action.connect("activate",
                             self.__on_radio_action_activate,
                             artist_ids)
        radio_action.set_enabled(get_network_available("SPOTIFY"))
        menu_item = Gio.MenuItem.new(_("Spotify"),
                                     "app.radio_action_spotify")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        section.append_item(menu_item)
        self.append_section(_("From the Web"), section)

#######################
# PRIVATE             #
#######################
    def __on_radio_action_activate(self, action, variant, artist_ids):
        """
            Play a radio from storage type
            @param Gio.SimpleAction
            @param GLib.Variant
            @param artist_ids as [int]
        """
        def play_radio():
            if action.get_name() == "radio_action_collection":
                App().player.play_radio_from_collection(artist_ids)
            elif action.get_name() == "radio_action_spotify":
                App().player.play_radio_from_spotify(artist_ids)
            elif action.get_name() == "radio_action_lastfm":
                App().player.play_radio_from_lastfm(artist_ids)
            elif action.get_name() == "radio_action_deezer":
                App().player.play_radio_from_deezer(artist_ids)
            elif action.get_name() == "radio_action_loved":
                App().player.play_radio_from_loved(artist_ids)
            elif action.get_name() == "radio_action_populars":
                App().player.play_radio_from_populars(artist_ids)
        App().task_helper.run(play_radio)


class PlaylistPlaybackMenu(Gio.Menu):
    """
        Contextual menu for a playlist
    """

    def __init__(self, playlist_id):
        """
            Init menu
            @param playlist id as int
        """
        Gio.Menu.__init__(self)
        play_action = Gio.SimpleAction(name="playlist_play_action")
        App().add_action(play_action)
        play_action.connect("activate", self.__on_play_action, playlist_id)
        menu_item = Gio.MenuItem.new(_("Play this playlist"),
                                     "app.playlist_play_action")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        self.append_item(menu_item)
        add_action = Gio.SimpleAction(name="playlist_add_action")
        App().add_action(add_action)
        add_action.connect("activate", self.__on_add_action, playlist_id)
        menu_item = Gio.MenuItem.new(_("Add this playlist"),
                                     "app.playlist_add_action")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        self.append_item(menu_item)

#######################
# PRIVATE             #
#######################
    def __get_albums(self, playlist_id):
        """
            Get albums for playlist_id
            @parma playlist_id as int
        """
        if App().playlists.get_smart(playlist_id):
            request = App().playlists.get_smart_sql(playlist_id)
            # We need to inject skipped/storage_type
            storage_type = get_default_storage_type()
            split = request.split("ORDER BY")
            split[0] += " AND loved != %s" % Type.NONE
            split[0] += " AND tracks.storage_type&%s " % storage_type
            track_ids = App().db.execute("ORDER BY".join(split))
            albums = tracks_to_albums(
                [Track(track_id) for track_id in track_ids])
        else:
            tracks = App().playlists.get_tracks(playlist_id)
            albums = tracks_to_albums(tracks)
        return albums

    def __on_play_action(self, action, variant, playlist_id):
        """
            Play albums
            @param Gio.SimpleAction
            @param GLib.Variant
            @param playlist_id as int
        """
        App().player.play_albums(self.__get_albums(playlist_id))

    def __on_add_action(self, action, variant, playlist_id):
        """
            Add albums to playback
            @param Gio.SimpleAction
            @param GLib.Variant
            @param playlist_id as int
        """
        App().player.add_albums(self.__get_albums(playlist_id))


class ArtistPlaybackMenu(PlaybackMenu):
    """
        Contextual menu for an artist
    """

    def __init__(self, artist_id, storage_type):
        """
            Init menu
            @param artist id as int
            @param storage_type as StorageType
        """
        PlaybackMenu.__init__(self)
        self.__artist_id = artist_id
        self.__storage_type = storage_type
        play_action = Gio.SimpleAction(name="artist_play_action")
        App().add_action(play_action)
        play_action.connect("activate", self.__on_play_action_activate)
        menu_item = Gio.MenuItem.new(_("Play this artist"),
                                     "app.artist_play_action")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        self.append_item(menu_item)
        self._set_playback_actions()
        self.__set_skipped_action()
        if get_network_available("SPOTIFY") or\
                get_network_available("LASTFM") or\
                get_network_available("DEEZER"):
            submenu = RadioPlaybackMenu([artist_id])
            self.append_submenu(_("Play a radio"), submenu)
        else:
            self._set_radio_action([artist_id])

    @property
    def in_player(self):
        """
            True if current object in player
            return bool
        """
        album_ids = App().albums.get_ids([], [self.__artist_id],
                                         self.__storage_type, False)
        return set(App().player.album_ids) & set(album_ids) == set(album_ids)

#######################
# PROTECTED           #
#######################
    def _append_to_playback(self, action, variant):
        """
            Append track to playback
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        from lollypop.utils_artist import add_artist_to_playback
        add_artist_to_playback([self.__artist_id], (), True)

    def _remove_from_playback(self, action, variant):
        """
            Delete track id from playback
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        from lollypop.utils_artist import add_artist_to_playback
        add_artist_to_playback([self.__artist_id], (), False)

#######################
# PRIVATE             #
#######################
    def __set_skipped_action(self):
        """
            Set skipped action
        """
        album_ids = App().albums.get_ids([], [self.__artist_id],
                                         self.__storage_type, False)
        skipped = True
        for album_id in album_ids:
            album = Album(album_id)
            if not album.loved & LovedFlags.SKIPPED:
                skipped = False
                break
        action = Gio.SimpleAction.new_stateful(
                "skip-artist",
                None,
                GLib.Variant.new_boolean(skipped))
        App().add_action(action)
        action.connect("change-state", self.__on_loved_change_state)
        self.append(_("Ignored"), "app.skip-artist")

    def __on_loved_change_state(self, action, state):
        """
            Update Skipped state
            @param action as Gio.SimpleAction
            @param state as bool
        """
        action.set_state(state)
        album_ids = App().albums.get_ids([], [self.__artist_id],
                                         self.__storage_type, False)
        for album_id in album_ids:
            album = Album(album_id)
            if state:
                album.set_loved(album.loved | LovedFlags.SKIPPED)
            else:
                album.set_loved(album.loved & ~LovedFlags.SKIPPED)

    def __on_play_action_activate(self, action, variant):
        """
            Play albums
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        from lollypop.utils_artist import play_artists
        play_artists([self.__artist_id], [])


class GenrePlaybackMenu(PlaybackMenu):
    """
        Contextual menu for a genre
    """

    def __init__(self, genre_id):
        """
            Init decade menu
            @param genre_id as int
        """
        PlaybackMenu.__init__(self)
        self.__genre_id = genre_id
        play_action = Gio.SimpleAction(name="genre_play_action")
        App().add_action(play_action)
        play_action.connect("activate", self.__play)
        menu_item = Gio.MenuItem.new(_("Play this genre"),
                                     "app.genre_play_action")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        self.append_item(menu_item)
        self._set_playback_actions()

    @property
    def in_player(self):
        """
            True if current object in player
            return bool
        """
        album_ids = self.__get_album_ids()
        return set(App().player.album_ids) & set(album_ids) == set(album_ids)

#######################
# PROTECTED           #
#######################
    def _append_to_playback(self, action, variant):
        """
            Append track to playback
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        album_ids = self.__get_album_ids()
        App().player.add_album_ids(album_ids)

    def _remove_from_playback(self, action, variant):
        """
            Delete track id from playback
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        album_ids = self.__get_album_ids()
        if App().player.current_track.album.id in album_ids:
            App().player.skip_album()
        App().player.remove_album_by_ids(album_ids)

#######################
# PRIVATE             #
#######################
    def __get_album_ids(self):
        """
            Get album ids for genre
            @return [int]
        """
        storage_type = get_default_storage_type()
        album_ids = App().albums.get_compilation_ids([self.__genre_id],
                                                     storage_type,
                                                     False)
        album_ids += App().albums.get_ids([self.__genre_id], [],
                                          storage_type,
                                          False)
        return album_ids

    def __play(self, action, variant):
        """
            Play albums
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        album_ids = self.__get_album_ids()
        albums = [Album(album_id) for album_id in album_ids]
        App().player.play_albums(albums)


class DecadePlaybackMenu(PlaybackMenu):
    """
        Contextual menu for a decade
    """

    def __init__(self, years):
        """
            Init decade menu
            @param years as [int]
        """
        PlaybackMenu.__init__(self)
        self.__years = years
        play_action = Gio.SimpleAction(name="decade_play_action")
        App().add_action(play_action)
        play_action.connect("activate", self.__play)
        menu_item = Gio.MenuItem.new(_("Play this decade"),
                                     "app.decade_play_action")
        menu_item.set_attribute_value("close", GLib.Variant("b", True))
        self.append_item(menu_item)
        self._set_playback_actions()

    @property
    def in_player(self):
        """
            True if current object in player
            return bool
        """
        album_ids = self.__get_album_ids()
        return set(App().player.album_ids) & set(album_ids) == set(album_ids)

#######################
# PROTECTED           #
#######################
    def _append_to_playback(self, action, variant):
        """
            Append track to playback
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        album_ids = self.__get_album_ids()
        App().player.add_album_ids(album_ids)

    def _remove_from_playback(self, action, variant):
        """
            Delete track id from playback
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        album_ids = self.__get_album_ids()
        if App().player.current_track.album.id in album_ids:
            App().player.skip_album()
        App().player.remove_album_by_ids(album_ids)

#######################
# PRIVATE             #
#######################
    def __get_album_ids(self):
        """
            Get album ids for decade
            @return [int]
        """
        storage_type = get_default_storage_type()
        album_ids = []
        for year in self.__years:
            album_ids += App().albums.get_compilation_ids_for_year(
                year, storage_type, False)
            album_ids += App().albums.get_ids_for_year(
                year, storage_type, False)
        return album_ids

    def __play(self, action, variant):
        """
            Play albums
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        album_ids = self.__get_album_ids()
        albums = [Album(album_id) for album_id in album_ids]
        App().player.play_albums(albums)


class AlbumPlaybackMenu(PlaybackMenu):
    """
        Contextual menu for an album
    """

    def __init__(self, album, view_type):
        """
            Init album menu
            @param album as Album
            @param view_type as ViewType
        """
        PlaybackMenu.__init__(self)
        self.__album = album
        if not view_type & ViewType.BANNER:
            play_action = Gio.SimpleAction(name="play_action")
            App().add_action(play_action)
            play_action.connect("activate", self.__on_play_action_activate)
            menu_item = Gio.MenuItem.new(_("Play this album"),
                                         "app.play_action")
            menu_item.set_attribute_value("close", GLib.Variant("b", True))
            self.append_item(menu_item)
            self._set_playback_actions()
        action = Gio.SimpleAction.new_stateful(
                "skip-album",
                None,
                GLib.Variant.new_boolean(
                    self.__album.loved & LovedFlags.SKIPPED))
        App().add_action(action)
        action.connect("change-state", self.__on_loved_change_state)
        self.append(_("Ignored"), "app.skip-album")
        if get_network_available("SPOTIFY") or\
                get_network_available("LASTFM") or\
                get_network_available("DEEZER"):
            submenu = RadioPlaybackMenu(album.artist_ids)
            self.append_submenu(_("Play a radio"), submenu)
        else:
            self._set_radio_action(album.artist_ids)

    @property
    def in_player(self):
        """
            True if current object in player
            return bool
        """
        return self.__album.id in App().player.album_ids

#######################
# PROTECTED           #
#######################
    def _append_to_playback(self, action, variant):
        """
            Append track to playback
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        App().player.add_album(self.__album)

    def _remove_from_playback(self, action, variant):
        """
            Delete track id from playback
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        if App().player.current_track.album.id == self.__album.id:
            App().player.skip_album()
        App().player.remove_album_by_id(self.__album.id)

#######################
# PRIVATE             #
#######################
    def __on_loved_change_state(self, action, state):
        """
            Update Skipped state
            @param action as Gio.SimpleAction
            @param state as bool
        """
        action.set_state(state)
        if state:
            self.__album.set_loved(self.__album.loved | LovedFlags.SKIPPED)
        else:
            self.__album.set_loved(self.__album.loved & ~LovedFlags.SKIPPED)

    def __on_play_action_activate(self, action, variant):
        """
            Play album
            @param action as Gio.SimpleAction
            @param variant as GLib.Variant
        """
        App().player.play_album(self.__album.clone(True))


class TrackPlaybackMenu(PlaybackMenu):
    """
        Contextual menu for tracks
    """

    def __init__(self, track, view_type):
        """
            Init track menu
            @param track as Track
            @param view_type as ViewType
        """
        PlaybackMenu.__init__(self)
        self.__track = track
        if not view_type & ViewType.TOOLBAR:
            self._set_playback_actions()
            self.__set_queue_actions()
        self.__set_stop_after_action()
        action = Gio.SimpleAction.new_stateful(
                "skip-track",
                None,
                GLib.Variant.new_boolean(
                    self.__track.loved & LovedFlags.SKIPPED))
        App().add_action(action)
        action.connect("change-state", self.__on_loved_change_state)
        self.append(_("Ignored"), "app.skip-track")

    @property
    def in_player(self):
        """
            True if current object in player
            return bool
        """
        for album in App().player.albums:
            if self.__track.album.id == album.id:
                if self.__track.id in album.track_ids:
                    return True
        return False

#######################
# PROTECTED           #
#######################
    def _append_to_playback(self, action, variant):
        """
            Append track to playback
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        album = Album(self.__track.album.id)
        album.set_tracks([self.__track])
        App().player.add_album(album)
        if App().player.is_playing:
            App().player.add_album(album)
        else:
            App().player.play_album(album)

    def _remove_from_playback(self, action, variant):
        """
            Delete track from playback
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        for album in App().player.albums:
            if album.id == self.__track.album.id:
                if self.__track.id in album.track_ids:
                    index = album.track_ids.index(self.__track.id)
                    track = album.tracks[index]
                    App().player.remove_track_from_album(track, album)
                    break

#######################
# PRIVATE             #
#######################
    def __set_queue_actions(self):
        """
            Set queue actions
        """
        if self.__track.id not in App().player.queue:
            append_queue_action = Gio.SimpleAction(name="append_queue_action")
            App().add_action(append_queue_action)
            append_queue_action.connect("activate",
                                        self.__append_to_queue)
            self.append(_("Add to queue"), "app.append_queue_action")
        else:
            del_queue_action = Gio.SimpleAction(name="del_queue_action")
            App().add_action(del_queue_action)
            del_queue_action.connect("activate",
                                     self.__remove_from_queue)
            self.append(_("Remove from queue"), "app.del_queue_action")

    def __set_stop_after_action(self):
        """
            Add an action to stop playback after track
        """
        if self.in_player and isinstance(self.__track, Track):
            stop_after_action = Gio.SimpleAction(name="stop_after_action")
            App().add_action(stop_after_action)
            if self.__track.id == App().player.stop_after_track_id:
                stop_after_action.connect("activate", self.__stop_after, None)
                self.append(_("Do not stop after"),
                            "app.stop_after_action")
            else:
                stop_after_action.connect("activate", self.__stop_after,
                                          self.__track.id)
                self.append(_("Stop after"), "app.stop_after_action")

    def __stop_after(self, action, variant, track_id):
        """
            Tell player to stop after track
            @param Gio.SimpleAction
            @param GLib.Variant
            @param track_id as int/None
        """
        App().player.stop_after(track_id)
        if track_id == App().player.current_track.id:
            App().player.set_next()

    def __append_to_queue(self, action, variant):
        """
            Append track to queue
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        App().player.append_to_queue(self.__track.id, False)
        emit_signal(App().player, "queue-changed")

    def __remove_from_queue(self, action, variant):
        """
            Delete track id from queue
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        App().player.remove_from_queue(self.__track.id, False)
        emit_signal(App().player, "queue-changed")

    def __on_loved_change_state(self, action, state):
        """
            Update Skipped state
            @param action as Gio.SimpleAction
            @param state as bool
        """
        action.set_state(state)
        if state:
            self.__track.set_loved(self.__track.loved | LovedFlags.SKIPPED)
        else:
            self.__track.set_loved(self.__track.loved & ~LovedFlags.SKIPPED)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk, Pango, GObject

from lollypop.define import ArtSize, ViewType, MARGIN, MARGIN_MEDIUM, App
from lollypop.utils import on_query_tooltip


class RoundedFlowBoxWidget(Gtk.FlowBoxChild):
    """
        Rounded flowbox child widget
    """

    __gsignals__ = {
        "populated": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, data, name, sortname, view_type, font_height):
        """
            Init widget
            @param data as object
            @param name as str
            @param sortname as str
            @param view_type as ViewType
            @param font_height as int
        """
        # We do not use Gtk.Builder for speed reasons
        Gtk.FlowBoxChild.__init__(self)
        self._artwork = None
        self._data = data
        self._art_size = 1
        self.__name = name
        self.__sortname = sortname
        self.__view_type = view_type
        self.__font_height = font_height
        self._scale_factor = self.get_scale_factor()
        self.set_property("halign", Gtk.Align.CENTER)
        self.set_property("margin", MARGIN)
        self.update_art_size()

    def populate(self):
        """
            Populate widget content
        """
        self._grid = Gtk.Grid()
        self._grid.set_orientation(Gtk.Orientation.VERTICAL)
        self._grid.set_row_spacing(MARGIN_MEDIUM)
        self._grid.show()
        self._label = Gtk.Label.new()
        self._label.set_ellipsize(Pango.EllipsizeMode.END)
        self._label.set_property("halign", Gtk.Align.CENTER)
        self._label.set_property("has-tooltip", True)
        self._label.connect("query-tooltip", on_query_tooltip)
        self._label.set_markup(
            "<b>" + GLib.markup_escape_text(self.__name) + "</b>")
        self._label.show()
        self._artwork = Gtk.Image.new()
        self._artwork.show()
        self._artwork.set_size_request(self._art_size, self._art_size)
        self.set_artwork()
        self._grid.add(self._artwork)
        self._grid.add(self._label)
        self.add(self._grid)

    def update_art_size(self):
        """
            Update art size based on current window state
        """
        if self.__view_type & ViewType.SMALL:
            self._art_size = ArtSize.MEDIUM
        elif App().window.folded:
            self._art_size = ArtSize.BANNER
        else:
            self._art_size = ArtSize.BIG
        self.set_size_request(self._art_size,
                              self._art_size + self.__font_height * 2)
        if self._artwork is not None:
            self._artwork.set_size_request(self._art_size, self._art_size)

    def set_artwork(self):
        """
            Set artwork
        """
        pass

    def reset_artwork(self):
        """
            Reset widget artwork
        """
        self.update_art_size()
        if self._artwork is not None:
            self._artwork.set_from_surface(None)

    def do_get_preferred_width(self):
        """
            Return preferred width
            @return (int, int)
        """
        width = Gtk.FlowBoxChild.do_get_preferred_width(self)[0]
        return (width, width)

    def rename(self, name):
        """
            Rename widget
            @param name as str
        """
        self._label.set_markup("<b>" + GLib.markup_escape_text(name) + "</b>")

    @property
    def name(self):
        """
            Get name
            @return str
        """
        return self.__name

    @property
    def sortname(self):
        """
            Get sortname
            @return str
        """
        return self.__sortname

    @property
    def data(self):
        """
            Get associated data
            @return object
        """
        return self._data

    @property
    def is_populated(self):
        """
            True if album populated
        """
        return True

    @property
    def artwork(self):
        """
            Get album artwork
            @return Gtk.Image
        """
        return self._artwork

    @property
    def artwork_name(self):
        """
            Get artwork name
            return str
        """
        return self.name

#######################
# PROTECTED           #
#######################

    def _get_album_ids(self):
        return []

#######################
# PRIVATE             #
#######################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject

from lollypop.define import StorageType, App
from lollypop.utils import emit_signal, get_network_available
from lollypop.search_local import LocalSearch


class Search(GObject.Object):
    """
        Local search
    """
    __gsignals__ = {
        "match-artist": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
        "match-album": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
        "match-track": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
        "finished": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    def __init__(self):
        """
            Init search
        """
        GObject.Object.__init__(self)
        self.__search_count = 0
        self.__local_search = LocalSearch()
        self.__connect_search_signals(self.__local_search)
        self.__web_search = None

    def set_web_search(self, name):
        """
            Set web search to name
            @param name as str
        """
        self.__web_search = None
        if not get_network_available("YOUTUBE"):
            return

        if name == "SPOTIFY":
            from lollypop.search_spotify import SpotifySearch
            self.__web_search = SpotifySearch()
            self.__connect_search_signals(self.__web_search)
        elif name == "LASTFM":
            from lollypop.search_lastfm import LastFMSearch
            self.__web_search = LastFMSearch()
            self.__connect_search_signals(self.__web_search)
        elif name == "MUSICBRAINZ":
            from lollypop.search_musicbrainz import MusicBrainzSearch
            self.__web_search = MusicBrainzSearch()
            self.__connect_search_signals(self.__web_search)
        elif name == "DEEZER":
            from lollypop.search_deezer import DeezerSearch
            self.__web_search = DeezerSearch()
            self.__connect_search_signals(self.__web_search)
        elif name == "JAMENDO":
            from lollypop.search_jamendo import JamendoSearch
            self.__web_search = JamendoSearch()
            self.__connect_search_signals(self.__web_search)

    def load_tracks(self, album, cancellable):
        """
            Load tracks for album with correct web service
            @param album as Album
            @param cancellable as Gio.Cancellable
        """
        uri = album.uri
        if not uri:
            pass
        elif uri.startswith("mb:"):
            from lollypop.search_musicbrainz import MusicBrainzSearch
            MusicBrainzSearch().load_tracks(album, cancellable)
        elif uri.startswith("dz:"):
            from lollypop.search_deezer import DeezerSearch
            DeezerSearch().load_tracks(album, cancellable)
        elif uri.startswith("jam:"):
            from lollypop.search_jamendo import JamendoSearch
            JamendoSearch().load_tracks(album, cancellable)
        # elif uri.startswith("sp:"): => compatibility
        else:
            from lollypop.search_spotify import SpotifySearch
            SpotifySearch().load_tracks(album, cancellable)

    def get(self, search, cancellable):
        """
            Get match for search
            @param search as str
            @param cancellable as Gio.Cancellable
        """
        # Only local items
        storage_type = StorageType.COLLECTION |\
            StorageType.SAVED |\
            StorageType.SEARCH
        App().task_helper.run(self.__local_search.get,
                              search, storage_type, cancellable)
        self.__search_count += 1
        if self.__web_search is not None:
            storage_type = StorageType.SEARCH | StorageType.EPHEMERAL
            App().task_helper.run(self.__web_search.get,
                                  search, storage_type, cancellable)
            self.__search_count += 1

#######################
# PRIVATE             #
#######################
    def __on_finished(self, search):
        """
            Emit finished signals if all search are finished
            @param search as Search provider
        """
        self.__search_count -= 1
        emit_signal(self, "finished", self.__search_count == 0)

    def __connect_search_signals(self, search):
        """
            Connect search signals
            @param search as Search provider
        """
        search.connect("match-artist",
                       lambda x, y, z: self.emit("match-artist", y, z))
        search.connect("match-album",
                       lambda x, y, z: self.emit("match-album", y, z))
        search.connect("match-track",
                       lambda x, y, z: self.emit("match-track", y, z))
        search.connect("finished", self.__on_finished)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from lollypop.define import App, ViewType, MARGIN, ScanUpdate
from lollypop.view_tracks_album import AlbumTracksView
from lollypop.widgets_banner_album import AlbumBannerWidget
from lollypop.view_lazyloading import LazyLoadingView
from lollypop.helper_signals import SignalsHelper, signals_map


class AlbumView(LazyLoadingView, SignalsHelper):
    """
        Show artist albums and tracks
    """

    @signals_map
    def __init__(self, album, storage_type, view_type):
        """
            Init ArtistView
            @param album as Album
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        LazyLoadingView.__init__(self, storage_type, view_type)
        self.__tracks_view = None
        self.__album = album
        self.__others_boxes = []
        self.__grid = Gtk.Grid()
        self.__grid.show()
        self.__grid.set_row_spacing(10)
        self.__grid.set_orientation(Gtk.Orientation.VERTICAL)
        self.__banner = AlbumBannerWidget(self.__album, self.storage_type,
                                          self.view_type)
        self.__banner.show()
        self.__banner.populate()
        self.add_widget(self.__grid, self.__banner)
        return [
            (App().scanner, "scan-finished", "_on_scan_finished"),
            (App().scanner, "updated", "_on_collection_updated"),
        ]

    def populate(self):
        """
            Populate the view with album
        """
        def do():
            if self.__tracks_view is None:
                self.__tracks_view = AlbumTracksView(self.__album,
                                                     self.view_type)
                self.__tracks_view.show()
                self.__tracks_view.connect("populated",
                                           self.__on_tracks_populated)
                self.__tracks_view.connect("selected",
                                           self.__on_track_selected)
                self.__tracks_view.set_margin_start(MARGIN)
                self.__tracks_view.set_margin_end(MARGIN)
                self.__tracks_view.populate()
                self.__grid.add(self.__tracks_view)
        # Let the banner populate first
        GLib.idle_add(do)

    @property
    def name(self):
        """
            Get name
            @return str
        """
        return self.__album.name

    @property
    def is_populated(self):
        """
            True if populated
            @return bool
        """
        return self.__tracks_view is not None

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"album": self.__album,
                "storage_type": self.storage_type,
                "view_type": self.view_type & ~ViewType.SMALL}

    @property
    def filtered(self):
        """
            Get filtered children
            @return [Gtk.Widget]
        """
        if self.__tracks_view is None:
            return []
        filtered = self.__tracks_view.children
        for box in self.__others_boxes:
            for child in box.children:
                filtered.append(child)
        return filtered

    @property
    def scroll_shift(self):
        """
            Get scroll shift for y axes
            @return int
        """
        return self.__banner.height

#######################
# PROTECTED           #
#######################
    def _on_scan_finished(self, scanner, track_ids):
        """
            Reload album if needed
            @param scanner as CollectionScanner
            @param track_ids as int
        """
        if not self.get_sensitive():
            App().window.container.reload_view()

    def _on_collection_updated(self, scanner, item, scan_update):
        """
            Handles changes in collection
            @param scanner as CollectionScanner
            @param item as CollectionItem
            @param scan_update as ScanUpdate
        """
        if item.album_id != self.__album.id:
            return
        if scan_update == ScanUpdate.REMOVED:
            App().window.container.go_back()

#######################
# PRIVATE             #
#######################
    def __on_tracks_populated(self, view):
        """
            Populate remaining discs
            @param view as TracksView
        """
        if self.__tracks_view.is_populated:
            self.emit("populated")
            if self.view_type & ViewType.OVERLAY:
                from lollypop.view_albums_line import AlbumsArtistLineView
                for artist_id in self.__album.artist_ids:
                    others_box = AlbumsArtistLineView(artist_id,
                                                      self.__album.genre_ids,
                                                      self.storage_type,
                                                      ViewType.SMALL |
                                                      ViewType.ALBUM |
                                                      ViewType.SCROLLED)
                    others_box.set_margin_start(MARGIN)
                    others_box.set_margin_end(MARGIN)
                    others_box.populate(self.__album.id)
                    self.__grid.add(others_box)
                    self.__others_boxes.append(others_box)
        else:
            GLib.idle_add(self.__tracks_view.populate)

    def __on_track_selected(self, view, selected):
        """
            Mark banner as filtered
            @param view as TracksView
            @param selected as bool
        """
        self.__banner.set_selected(selected)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from random import shuffle

from lollypop.utils import get_human_duration, popup_widget
from lollypop.utils_album import tracks_to_albums
from lollypop.define import App, ArtSize, ViewType
from lollypop.widgets_banner import BannerWidget
from lollypop.helper_signals import SignalsHelper, signals_map


class PlaylistBannerWidget(BannerWidget, SignalsHelper):
    """
        Banner for playlist
    """

    @signals_map
    def __init__(self, playlist_id, view):
        """
            Init banner
            @param playlist_id as int
            @param view as AlbumsListView
        """
        BannerWidget.__init__(self, view.args["view_type"] | ViewType.OVERLAY)
        self.__duration_task = False
        self.__playlist_id = playlist_id
        self.__view = view
        builder = Gtk.Builder()
        builder.add_from_resource(
            "/org/gnome/Lollypop/PlaylistBannerWidget.ui")
        self.__title_label = builder.get_object("title")
        self.__spinner = builder.get_object("spinner")
        self.__duration_label = builder.get_object("duration")
        self.__play_button = builder.get_object("play_button")
        self.__shuffle_button = builder.get_object("shuffle_button")
        self.__menu_button = builder.get_object("menu_button")
        widget = builder.get_object("widget")
        self._overlay.add_overlay(widget)
        self._overlay.set_overlay_pass_through(widget, True)
        self.__title_label.set_label(App().playlists.get_name(playlist_id))
        builder.connect_signals(self)
        self.connect("destroy", self.__on_destroy)
        self.__set_internal_size()
        return [
            (view, "initialized", "_on_view_updated"),
            (App().window.container.widget, "notify::folded",
             "_on_container_folded"),
            (App().player, "duration-changed", "_on_view_updated"),
            (App().playlists, "playlist-track-added", "_on_playlist_changed"),
            (App().playlists, "playlist-track-removed",
             "_on_playlist_changed"),
        ]

    def rename(self, name):
        """
            Rename playlist
            @param name as str
        """
        self.__title_label.set_label(name)

    @property
    def spinner(self):
        """
            Get spinner
            @return Gtk.Spinner
        """
        return self.__spinner

    @property
    def height(self):
        """
            Get wanted height
            @return int
        """
        return ArtSize.SMALL

#######################
# PROTECTED           #
#######################
    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        self.__set_internal_size()

    def _on_view_updated(self, *ignore):
        """
            Update duration
        """
        if self.__view.children:
            self.__duration_task = True
            App().task_helper.run(self.__calculate_duration)
        else:
            self.__duration_task = False
            self.__duration_label.set_text(get_human_duration(0))

    def _on_play_button_clicked(self, button):
        """
            Play playlist
            @param button as Gtk.Button
        """
        tracks = []
        for album_row in self.__view.children:
            for track in album_row.album.tracks:
                tracks.append(track)
        if tracks:
            albums = tracks_to_albums(tracks)
            App().player.play_track_for_albums(tracks[0], albums)

    def _on_shuffle_button_clicked(self, button):
        """
            Play playlist shuffled
            @param button as Gtk.Button
        """
        tracks = []
        for album_row in self.__view.children:
            for track in album_row.album.tracks:
                tracks.append(track)
        if tracks:
            shuffle(tracks)
            albums = tracks_to_albums(tracks)
            App().player.play_track_for_albums(tracks[0], albums)

    def _on_menu_button_clicked(self, button):
        """
            Show playlist menu
            @param button as Gtk.Button
        """
        from lollypop.widgets_menu import MenuBuilder
        from lollypop.menu_playlist import PlaylistMenu, PlaylistMenuExt
        menu = PlaylistMenu(self.__playlist_id,
                            self.view_type,
                            App().window.folded)
        menu_widget = MenuBuilder(menu)
        if self.__playlist_id >= 0:
            menu_widget = MenuBuilder(menu)
            menu_ext = PlaylistMenuExt(self.__playlist_id)
            menu_ext.show()
            menu_widget.add_widget(menu_ext)
        else:
            menu_widget = MenuBuilder(menu)
        menu_widget.show()
        popup_widget(menu_widget, button, None, None, button)

    def _on_playlist_changed(self, playlists, playlist_id, uri):
        """
            Update duration
            @param playlists as Playlists
            @param playlist_id as int
            @param uri as str
        """
        if self.__playlist_id == playlist_id:
            self.__calculate_duration()

#######################
# PRIVATE             #
#######################
    def __calculate_duration(self):
        """
            Calculate playback duration
        """
        duration = 0
        for child in self.__view.children:
            if not self.__duration_task:
                return
            duration += child.album.duration
        GLib.idle_add(self.__duration_label.set_text,
                      get_human_duration(duration))

    def __set_internal_size(self):
        """
            Update font size
        """
        duration_context = self.__duration_label.get_style_context()
        title_context = self.__title_label.get_style_context()
        for c in title_context.list_classes():
            title_context.remove_class(c)
        for c in duration_context.list_classes():
            duration_context.remove_class(c)
        if App().window.folded:
            title_context.add_class("text-large")
        else:
            title_context.add_class("text-x-large")
            duration_context.add_class("text-large")

    def __on_get_duration(self, duration):
        """
            Update label
            @param duration as int
        """
        self.__duration_label.set_text(get_human_duration(duration))

    def __on_destroy(self, widget):
        """
            Remove ref cycle
            @param widget as Gtk.Widget
        """
        self.__view = None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gio, GLib

from gettext import gettext as _

from lollypop.pop_devices import DevicesPopover
from lollypop.define import App, Repeat, Type
from lollypop.utils import popup_widget, emit_signal, get_network_available
from lollypop.progressbar import ButtonProgressBar


class ToolbarEnd(Gtk.Bin):
    """
        Toolbar end
    """

    def __init__(self, window):
        """
            Init toolbar
            @param window as Window
        """
        Gtk.Bin.__init__(self)
        self.set_hexpand(True)
        self.__search_popover = None
        self.__devices_popover = None
        self.__app_menu = None
        self.__playback_menu = None
        self.__timeout_id = None
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/ToolbarEnd.ui")
        self.__shuffle_menu = builder.get_object("shuffle-menu")
        self.__party_submenu = builder.get_object("party_submenu")
        self.add(builder.get_object("end"))

        self.__shuffle_action = Gio.SimpleAction.new_stateful(
            "shuffle",
            None,
            App().settings.get_value("shuffle"))
        self.__shuffle_action.connect("change-state",
                                      self.__on_shuffle_change_state)
        self.__repeat_action = Gio.SimpleAction.new_stateful(
            "repeat",
            GLib.VariantType.new("s"),
            GLib.Variant("s", "none"))
        self.__repeat_action.set_state(App().settings.get_value("repeat"))
        self.__repeat_action.connect("change-state",
                                     self.__on_repeat_change_state)
        App().add_action(self.__shuffle_action)
        App().add_action(self.__repeat_action)

        self.__playback_button = builder.get_object("playback_button")
        self.__playback_button_image = builder.get_object(
            "shuffle_button_image")
        shuffle_button_action = Gio.SimpleAction.new("shuffle-button", None)
        shuffle_button_action.connect("activate",
                                      self.__on_shuffle_button_activate)
        App().add_action(shuffle_button_action)
        App().set_accels_for_action("app.shuffle-button", ["<Control>r"])
        App().settings.connect("changed::shuffle", self.__on_repeat_changed)
        App().settings.connect("changed::repeat", self.__on_repeat_changed)

        party_action = Gio.SimpleAction.new_stateful(
            "party",
            None,
            GLib.Variant.new_boolean(App().player.is_party))
        party_action.connect("change-state", self.__on_party_mode_change_state)
        App().add_action(party_action)
        App().set_accels_for_action("app.party", ["<Control>p"])

        scrobbling_disabled = App().settings.get_value("disable-scrobbling")
        scrobbling_action = Gio.SimpleAction.new_stateful(
            "scrobbling",
            None,
            GLib.Variant.new_boolean(not scrobbling_disabled))
        scrobbling_action.connect("change-state",
                                  self.__on_scrobbling_mode_change_state)
        App().add_action(scrobbling_action)
        App().set_accels_for_action("app.scrobbling", ["<Control><Shift>s"])

        self.__menu_button = builder.get_object("menu_button")
        self.__home_button = builder.get_object("home_button")
        self.__devices_button = builder.get_object("devices_button")
        self.__playback_button = builder.get_object("playback_button")
        self.__set_shuffle_icon()

        button_progress_bar = ButtonProgressBar()
        overlay = builder.get_object("overlay")
        overlay.add_overlay(button_progress_bar)
        overlay.set_overlay_pass_through(button_progress_bar, True)
        self.__devices_popover = DevicesPopover(button_progress_bar)
        self.__devices_popover.connect(
                "hidden", self.__on_menu_hidden, self.__devices_button)
        self.__devices_popover.connect("content-changed",
                                       self.__on_devices_content_changed)
        self.__devices_popover.populate()
        builder.connect_signals(self)
        window.container.widget.connect("notify::folded",
                                        self.__on_container_folded)
        window.container.connect("can-go-back-changed",
                                 self.__on_can_go_back_changed)
        self.__on_container_folded(None, "folded")

    @property
    def devices_popover(self):
        """
            Get Devices Popover
            @return DevicesPopover
        """
        return self.__devices_popover

#######################
# PROTECTED           #
#######################
    def _on_home_button_clicked(self, button):
        """
            Go home in adaptive mode
            @param button as Gtk.Button
        """
        App().window.container.go_home()

    def _on_shuffle_button_toggled(self, button):
        """
           Popup shuffle menu
           @param button as Gtk.ToggleButton
        """
        if button.get_active():
            if self.__app_menu is not None:
                emit_signal(self.__app_menu, "hidden", True)
            self.__party_submenu.remove_all()
            self.__init_party_submenu()
            from lollypop.widgets_menu import MenuBuilder
            self.__playback_menu = MenuBuilder(self.__shuffle_menu)
            self.__playback_menu.show()
            popover = popup_widget(self.__playback_menu, button,
                                   None, None, None)
            if popover is None:
                self.__playback_menu.connect("hidden",
                                             self.__on_menu_hidden,
                                             button)
            else:
                popover.connect("hidden", self.__on_menu_hidden, button)
        elif self.__playback_menu is not None and App().window.folded:
            emit_signal(self.__playback_menu, "hidden", True)

    def _on_devices_button_toggled(self, button):
        """
           Create submenu
           @param button as Gtk.ToggleButton
        """
        if button.get_active():
            self.__devices_popover.set_relative_to(button)
            self.__devices_popover.popup()

    def _on_settings_button_toggled(self, button):
        """
           Popup application menu
           @param button as Gtk.ToggleButton
        """
        from lollypop.menu_application import ApplicationMenu
        if button.get_active():
            if self.__playback_menu is not None:
                emit_signal(self.__playback_menu, "hidden", True)
            self.__app_menu = ApplicationMenu()
            self.__app_menu.show()
            popover = popup_widget(self.__app_menu, button, None, None, None)
            if popover is None:
                self.__app_menu.connect("hidden",
                                        self.__on_menu_hidden,
                                        button)
            else:
                popover.connect("hidden", self.__on_menu_hidden, button)
        elif self.__app_menu is not None and App().window.folded:
            emit_signal(self.__app_menu, "hidden", True)

#######################
# PRIVATE             #
#######################
    def __init_party_submenu(self):
        """
            Init party submenu with current ids
        """
        def update_actions():
            # Update state with a thread
            party_ids = list(App().settings.get_value("party-ids"))
            all_ids = App().genres.get_ids() + [Type.WEB]
            all_selected = not party_ids or len(party_ids) == len(all_ids)
            App().lookup_action("all_party_ids").set_state(
                GLib.Variant.new_boolean(all_selected))
            for genre_id in all_ids:
                in_party_ids = not party_ids or genre_id in party_ids
                action = App().lookup_action("genre_%s" % genre_id)
                action.set_state(GLib.Variant.new_boolean(in_party_ids))

        def update_party_ids(party_ids, value, genre_id):
            all_ids = App().genres.get_ids() + [Type.WEB]
            if value:
                if genre_id not in party_ids:
                    party_ids.append(genre_id)
                if len(party_ids) == len(all_ids):
                    party_ids = []
            else:
                if not party_ids:
                    party_ids = all_ids
                if genre_id in party_ids:
                    party_ids.remove(genre_id)
            return party_ids

        def on_change_state(action, value, genre_id):
            party_ids = list(App().settings.get_value("party-ids"))
            all_ids = App().genres.get_ids() + [Type.WEB]
            action.set_state(value)
            if genre_id is None:
                for genre_id in all_ids:
                    action = App().lookup_action("genre_%s" % genre_id)
                    action.set_state(value)
                    party_ids = update_party_ids(party_ids, value, genre_id)
            else:
                party_ids = update_party_ids(party_ids, value, genre_id)
            all_selected = not party_ids or len(party_ids) == len(all_ids)
            App().lookup_action("all_party_ids").set_state(
                GLib.Variant("b", all_selected))
            App().settings.set_value("party-ids",
                                     GLib.Variant("ai", party_ids))
            App().task_helper.run(App().player.set_party_ids,
                                  callback=(
                                    lambda x: App().player.set_next(),))

        # Create actions
        action = Gio.SimpleAction.new_stateful(
                    "all_party_ids",
                    None,
                    GLib.Variant.new_boolean(False))
        action.connect("change-state", on_change_state, None)
        App().add_action(action)
        item = Gio.MenuItem.new(_("All genres"), "app.all_party_ids")
        self.__party_submenu.append_item(item)
        genres = App().genres.get()
        genres += [(Type.WEB, _("Web"), _("Web"))]
        for (genre_id, name, sortname) in genres:
            action_name = "genre_%s" % genre_id
            action = Gio.SimpleAction.new_stateful(
                action_name,
                None,
                GLib.Variant.new_boolean(False))
            action.connect("change-state", on_change_state, genre_id)
            if genre_id == Type.WEB:
                action.set_enabled(get_network_available("YOUTUBE"))
            App().add_action(action)
            menu_str = name if len(name) < 20 else name[0:20] + "…"
            item = Gio.MenuItem.new(menu_str, "app.%s" % action_name)
            self.__party_submenu.append_item(item)
        App().task_helper.run(update_actions)

    def __set_shuffle_icon(self):
        """
            Set shuffle icon
        """
        shuffle = App().settings.get_value("shuffle")
        repeat = App().settings.get_enum("repeat")
        if shuffle:
            self.__playback_button_image.set_from_icon_name(
                "media-playlist-shuffle-symbolic",
                Gtk.IconSize.BUTTON)
        elif repeat == Repeat.TRACK:
            self.__playback_button_image.get_style_context().remove_class(
                "selected")
            self.__playback_button_image.set_from_icon_name(
                "media-playlist-repeat-song-symbolic",
                Gtk.IconSize.BUTTON)
        elif repeat == Repeat.ALL:
            self.__playback_button_image.set_from_icon_name(
                "media-playlist-repeat-symbolic",
                Gtk.IconSize.BUTTON)
        else:
            self.__playback_button_image.set_from_icon_name(
                "media-playlist-consecutive-symbolic",
                Gtk.IconSize.BUTTON)

    def __on_party_mode_change_state(self, action, value):
        """
            Activate party mode
            @param action as Gio.SimpleAction
            @param value as bool
        """
        if not App().gtk_application_prefer_dark_theme and\
                not App().settings.get_value("dark-ui"):
            settings = Gtk.Settings.get_default()
            settings.set_property("gtk-application-prefer-dark-theme", value)
        App().player.set_party(value.get_boolean())
        action.set_state(value)
        self.__shuffle_action.set_enabled(not value)
        self.__repeat_action.set_enabled(not value)

    def __on_scrobbling_mode_change_state(self, action, value):
        """
            Change scrobbling option
            @param action as Gio.SimpleAction
            @param value as bool
        """
        action.set_state(value)
        App().settings.set_value("disable-scrobbling",
                                 GLib.Variant("b", not value))

    def __on_shuffle_change_state(self, action, value):
        """
            Update shuffle setting
            @param action as Gio.SimpleAction
            @param value as bool
        """
        App().settings.set_value("shuffle", value)
        action.set_state(value)

    def __on_repeat_change_state(self, action, value):
        """
            Update playback setting
            @param action as Gio.SimpleAction
            @param value as GLib.Variant
        """
        App().settings.set_value("repeat", value)
        action.set_state(value)

    def __on_shuffle_button_activate(self, action, param):
        """
            Activate shuffle button
            @param action as Gio.SimpleAction
            @param param as GLib.Variant
        """
        self.__playback_button.set_active(
            not self.__playback_button.get_active())

    def __on_repeat_changed(self, settings, value):
        """
            Update shuffle icon
            @param settings as Gio.Settings
            @param value as GLib.Variant
        """
        self.__set_shuffle_icon()

    def __on_menu_hidden(self, menu, hide, button):
        """
            Restore button state and reset menus
            @param menu as MenuWidget
            @param hide as bool
            @param button as Gtk.Button
        """
        self.__app_menu = None
        self.__playback_menu = None
        button.set_active(False)

    def __on_devices_content_changed(self, popover, count):
        """
            Show/Hide device button
            @param popover as DevicesPopover
            @param count as int
        """
        if count:
            self.__devices_button.show()
        else:
            self.__devices_button.hide()

    def __on_container_folded(self, leaflet, folded):
        """
            Show/Hide home button
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        if App().window.folded:
            self.__home_button.show()
            self.__home_button.set_sensitive(True)
        else:
            self.__home_button.hide()

    def __on_can_go_back_changed(self, container, back):
        """
            Make button sensitive
            @param container as Container
            @param back as bool
        """
        if back:
            self.__home_button.set_sensitive(True)
        else:
            self.__home_button.set_sensitive(False)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.view import View
from lollypop.define import ViewType, StorageType


class MenuView(View):
    """
        Show a menu
    """

    def __init__(self, menu):
        """
            Init view
            @param menu as Gtk.Widget
        """
        View.__init__(self, StorageType.ALL, ViewType.SCROLLED)
        menu.get_style_context().add_class("adaptive-menu")
        menu.set_vexpand(True)
        self.add_widget(menu)

    @property
    def args(self):
        return None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from itertools import chain

from lollypop.define import Repeat, App
from lollypop.objects_track import Track
from lollypop.logger import Logger


class LinearPlayer:
    """
        Manage normal playback
    """

    def __init__(self):
        """
            Init linear player
        """
        pass

    def next(self):
        """
            Next track for current album or next album
            @return track as Track
        """
        # No album in playback
        if not self.albums:
            return Track()
        album = self._current_track.album
        track = self.__fallback_track_if_album_missing(album)
        # Current album missing, go to fallback track
        if track is not None:
            return track
        new_track_position = self._current_track.position + 1
        # next album
        if new_track_position >= len(album.track_ids):
            try:
                pos = self.albums.index(album)
                albums_count = len(self._albums)
                new_pos = 0
                # Search for a next album
                for idx in chain(range(pos + 1, albums_count),
                                 range(0, pos)):
                    if self._albums[idx].tracks:
                        new_pos = idx
                        break
                if new_pos == 0:
                    repeat = App().settings.get_enum("repeat")
                    if repeat == Repeat.ALL:
                        pos = 0
                    else:
                        return Track()
                else:
                    pos = new_pos
            except Exception as e:
                Logger.error("LinearPlayer::next(): %s", e)
                pos = 0  # Happens if current album has been removed
            track = self._albums[pos].tracks[0]
        # next track
        else:
            track = album.tracks[new_track_position]
        return track

    def prev(self):
        """
            Prev track base on.current_track context
            @return track as Track
        """
        # No album in playback
        if not self._albums:
            return Track()
        album = self._current_track.album
        track = self.__fallback_track_if_album_missing(album)
        # Current album missing, go to fallback track
        if track is not None:
            return track
        new_track_position = self._current_track.position - 1
        # Previous album
        if new_track_position < 0:
            try:
                pos = self.albums.index(album)
                albums_count = len(self._albums)
                new_pos = 0
                # Search for a prev album
                for idx in chain(reversed(range(0, pos)),
                                 reversed(range(pos, albums_count))):
                    if self._albums[idx].tracks:
                        new_pos = idx
                        break
                if new_pos == albums_count - 1:
                    repeat = App().settings.get_enum("repeat")
                    if repeat == Repeat.ALL:
                        pos = new_pos
                    else:
                        return Track()
                else:
                    pos = new_pos
            except Exception as e:
                Logger.error("LinearPlayer::prev(): %s", e)
                pos = 0  # Happens if current album has been removed
            track = self._albums[pos].tracks[-1]
        # Previous track
        else:
            track = album.tracks[new_track_position]
        return track

    def __fallback_track_if_album_missing(self, album):
        """
            Get a fallback track if album not in player
            @param album as Album
            @return Track/None
        """
        if album not in self._albums:
            album = self._albums[0]
            if album.tracks:
                return album.tracks[0]
            else:
                return Track()
        return None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject, GLib

from collections import Counter

from lollypop.define import App
from lollypop.utils import noaccents


class LocalSearch(GObject.Object):
    """
        Local search
    """
    __gsignals__ = {
        "match-artist": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
        "match-album": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
        "match-track": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
        "finished": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self):
        """
            Init search
        """
        GObject.Object.__init__(self)

    def get(self, search, storage_type, cancellable):
        """
            Get match for search
            @param search as str
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
        """
        search = noaccents(search)
        self.__get_artists(search, storage_type, cancellable)
        self.__get_albums(search, storage_type, cancellable)
        self.__get_tracks(search, storage_type, cancellable)
        GLib.idle_add(self.emit, "finished")

#######################
# PRIVATE             #
#######################
    def __split_string(self, string):
        """
            Split string for search
            @param string as str
            @return str
        """
        split = []
        for word in string.split():
            if len(word) > 2:
                split.append(word)
        return split

    def __search_tracks(self, search, storage_type, cancellable):
        """
            Get tracks for search items
            @param search as str
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
            @return [int]
        """
        tracks = []
        track_ids = []
        split = self.__split_string(search)
        for search_str in [search] + split:
            tracks += App().tracks.search_performed(search_str, storage_type)
            tracks += App().tracks.search(search_str, storage_type)
            if cancellable.is_cancelled():
                break
        for (track_id, track_name) in tracks:
            valid = True
            no_accents = noaccents(track_name)
            if not no_accents.startswith(search):
                for word in split:
                    if word not in no_accents:
                        valid = False
                        break
            # Start with same word, adding to result
            else:
                track_ids.append(track_id)
            # All words are valid, adding to result
            if valid:
                track_ids.append(track_id)
            # Detect an artist match, adding to result
            for artist in App().tracks.get_artists(track_id):
                no_accents = noaccents(artist)
                for word in split:
                    if word in no_accents:
                        track_ids.append(track_id)
        return track_ids

    def __search_artists(self, search, storage_type, cancellable):
        """
            Get artists for search items
            @param search as str
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
            @return [int]
        """
        artists = []
        artist_ids = []
        split = self.__split_string(search)
        for search_str in [search] + split:
            artists += App().artists.search(search_str, storage_type)
            if cancellable.is_cancelled():
                break
        for (artist_id, artist_name) in artists:
            valid = True
            no_accents = noaccents(artist_name)
            if not no_accents.startswith(search):
                for word in split:
                    if word not in no_accents:
                        valid = False
                        break
            else:
                artist_ids.append(artist_id)
            if valid:
                artist_ids.append(artist_id)
        return artist_ids

    def __search_albums(self, search, storage_type, cancellable):
        """
            Get albums for search items
            @param search as str
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
            @return [int]
        """
        albums = []
        album_ids = []
        split = self.__split_string(search)
        for search_str in [search] + split:
            albums += App().albums.search(search_str, storage_type)
            if cancellable.is_cancelled():
                break
        for (album_id, album_name) in albums:
            valid = True
            no_accents = noaccents(album_name)
            if not no_accents.startswith(search):
                for word in split:
                    if word not in no_accents:
                        valid = False
                        break
            else:
                album_ids.append(album_id)
            if valid:
                album_ids.append(album_id)
        return album_ids

    def __get_artists(self, search, storage_type, cancellable):
        """
            Get artists for search
            @param search as str
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
        """
        artist_ids = self.__search_artists(search, storage_type, cancellable)
        counter = Counter(artist_ids)
        artist_ids = sorted(artist_ids,
                            key=lambda x: (counter[x], x),
                            reverse=True)
        artist_ids = list(dict.fromkeys(artist_ids))
        for artist_id in artist_ids:
            GLib.idle_add(self.emit, "match-artist", artist_id, storage_type)

    def __get_albums(self, search, storage_type, cancellable):
        """
            Get albums for search
            @param search as str
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
        """
        album_ids = self.__search_albums(search, storage_type, cancellable)
        counter = Counter(album_ids)
        album_ids = sorted(album_ids,
                           key=lambda x: (counter[x], x),
                           reverse=True)
        album_ids = list(dict.fromkeys(album_ids))
        for album_id in album_ids:
            GLib.idle_add(self.emit, "match-album", album_id, storage_type)

    def __get_tracks(self, search, storage_type, cancellable):
        """
            Get tracks for search
            @param search as str
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
        """
        track_ids = self.__search_tracks(search, storage_type, cancellable)
        counter = Counter(track_ids)
        track_ids = sorted(track_ids,
                           key=lambda x: (counter[x], x),
                           reverse=True)
        track_ids = list(dict.fromkeys(track_ids))
        for track_id in track_ids:
            GLib.idle_add(self.emit, "match-track", track_id, storage_type)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib, Gdk

from urllib.parse import urlparse

from lollypop.utils import emit_signal
from lollypop.utils_file import get_file_type
from lollypop.define import ScanType, FileType
from lollypop.logger import Logger
from lollypop.objects_track import Track
from lollypop.objects_album import Album


class ApplicationCmdline:
    """
        Handle command line switches
        Need to be inherited by a Gtk.Application
    """

    def __init__(self, version):
        """
            Create cmdline handler
            @param version as str
        """
        self.__debug = False
        self.__version = version
        self.add_main_option("play-ids", b"a", GLib.OptionFlags.NONE,
                             GLib.OptionArg.STRING, "Play ids", None)
        self.add_main_option("debug", b"d", GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Debug Lollypop", None)
        self.add_main_option("set-rating", b"r", GLib.OptionFlags.NONE,
                             GLib.OptionArg.STRING, "Rate the current track",
                             None)
        self.add_main_option("play-pause", b"t", GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Toggle playback",
                             None)
        self.add_main_option("stop", b"s", GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Stop playback",
                             None)
        self.add_main_option("next", b"n", GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Go to next track",
                             None)
        self.add_main_option("prev", b"p", GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Go to prev track",
                             None)
        self.add_main_option("emulate-phone", b"e", GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE,
                             "Emulate a Librem phone",
                             None)
        self.add_main_option("version", b"v", GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE,
                             "Lollypop version",
                             None)
        self.connect("command-line", self.__on_command_line)
        self.connect("handle-local-options", self.__on_handle_local_options)

    @property
    def version(self):
        """
            Get Lollypop version
            @return srt
        """
        return self.__version

    @property
    def debug(self):
        """
            Get debug state
            @return bool
        """
        return self.__debug

#######################
# PRIVATE             #
#######################
    def __parse_uris(self, playlist_uris, audio_uris):
        """
            Parse playlist uris
            @param playlist_uris as [str]
            @param audio_uris as [str]
        """
        from gi.repository import TotemPlParser
        playlist_uri = playlist_uris.pop(0)
        parser = TotemPlParser.Parser.new()
        parser.connect("entry-parsed",
                       self.__on_entry_parsed,
                       audio_uris)
        parser.parse_async(playlist_uri, True, None,
                           self.__on_parse_finished,
                           playlist_uris, audio_uris)

    def __on_handle_local_options(self, app, options):
        """
            Handle local options
            @param app as Gio.Application
            @param options as GLib.VariantDict
        """
        if options.contains("version"):
            print(self.__version)
            exit(0)
        self.register(None)
        if self.get_is_remote():
            Gdk.notify_startup_complete()
        return -1

    def __on_command_line(self, app, app_cmd_line):
        """
            Handle command line
            @param app as Gio.Application
            @param options as Gio.ApplicationCommandLine
        """
        try:
            args = app_cmd_line.get_arguments()
            options = app_cmd_line.get_options_dict()
            if options.contains("debug"):
                self.__debug = True
            if options.contains("set-rating"):
                value = options.lookup_value("set-rating").get_string()
                try:
                    value = min(max(0, int(value)), 5)
                    if self.player.current_track.id is not None:
                        self.player.current_track.set_rate(value)
                except Exception as e:
                    Logger.error("Application::__on_command_line(): %s", e)
                    pass
            elif options.contains("play-pause"):
                self.player.play_pause()
            elif options.contains("stop"):
                self.player.stop()
            elif options.contains("play-ids"):
                try:
                    value = options.lookup_value("play-ids").get_string()
                    ids = value.split(";")
                    albums = []
                    for id in ids:
                        if id[0:2] == "a:":
                            album = Album(int(id[2:]))
                            self.player.add_album(album)
                            albums.append(album)
                        else:
                            track = Track(int(id[2:]))
                            self.player.add_album(track.album)
                            albums.append(track.album)
                    if albums and albums[0].tracks:
                        self.player.load(albums[0].tracks[0])
                except Exception as e:
                    Logger.error("Application::__on_command_line(): %s", e)
                    pass
            elif options.contains("next"):
                self.player.next()
            elif options.contains("prev"):
                self.player.prev()
            elif options.contains("emulate-phone"):
                self.window.toolbar.end.devices_popover.add_fake_phone()
            elif len(args) > 1:
                audio_uris = []
                playlist_uris = []
                for uri in args[1:]:
                    parsed = urlparse(uri)
                    if parsed.scheme not in ["http", "https"]:
                        try:
                            uri = GLib.filename_to_uri(uri)
                        except:
                            pass
                        f = Gio.File.new_for_uri(uri)
                        # Try ./filename
                        if not f.query_exists():
                            uri = GLib.filename_to_uri(
                                "%s/%s" % (GLib.get_current_dir(), uri))
                            f = Gio.File.new_for_uri(uri)
                    file_type = get_file_type(uri)
                    if file_type == FileType.PLS:
                        playlist_uris.append(uri)
                    else:
                        audio_uris.append(uri)
                if playlist_uris:
                    self.__parse_uris(playlist_uris, audio_uris)
                else:
                    self.__on_parse_finished(None, None, [], audio_uris)
            elif self.window is not None:
                if not self.window.is_visible():
                    self.window.present()
                    emit_signal(self.player, "status-changed")
                    emit_signal(self.player, "current-changed")
            Gdk.notify_startup_complete()
        except Exception as e:
            Logger.error("Application::__on_command_line(): %s", e)
        return 0

    def __on_parse_finished(self, source, result, playlist_uris, audio_uris):
        """
            Play stream
            @param source as None
            @param result as Gio.AsyncResult
            @param uris as ([str], [str])
        """
        if playlist_uris:
            self.__parse_uris(playlist_uris, audio_uris)
        else:
            self.scanner.update(ScanType.EXTERNAL, audio_uris)

    def __on_entry_parsed(self, parser, uri, metadata, audio_uris):
        """
            Add playlist entry to external files
            @param parser as TotemPlParser.Parser
            @param uri as str
            @param metadata as GLib.HastTable
            @param audio_uris as str
        """
        audio_uris.append(uri)
